#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['RunCoro', 'arun_coro_fast', 'run_coro_fast', 'arun_coro', 'run_coro']

from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.introspection.inspect import get_exception, get_exception_tripple
from cengal.parallel_execution.coroutines.coro_standard_services_internal_lib.service_with_a_direct_request import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import put_current_from_other_service
from cengal.code_flow_control.smart_values import ValueExistence
from typing import Any, Optional, Tuple, Dict, Set, Union, List
import sys


class RunCoro(ServiceWithADirectRequestMixin, TypedService[Any]):
    def __init__(self, loop: CoroSchedulerType):
        super(RunCoro, self).__init__(loop)
        self.called_by: Dict[CoroID, CoroID] = dict()  # Dict[CoroID, CoroID] # key - callable; value - requester
        self.results: List[Tuple[CoroID, Any, Optional[BaseException]]] = list()  # (id, result, exception)
        self.results: Dict[CoroID, Tuple[Any, Optional[BaseException]]] = dict()  # (id, result, exception)
    
    def single_task_registration_or_immediate_processing(
            self, coro_worker: AnyWorker, *args, **kwargs) -> Tuple[bool, Any, Optional[BaseException]]:
        requester_id = self.current_caller_coro_info.coro_id
        try:
            coro: CoroWrapperBase = put_current_from_other_service(self, coro_worker, *args, **kwargs)
        except:
            ex_type, exception, tracback = get_exception_tripple()
            if __debug__: dlog(ex_type, exception, tracback)
            exception = exception.with_traceback(tracback)
            return True, None, exception
        
        coro.add_on_coro_del_handler(self._on_coro_del_handler)
        self.called_by[coro.coro_id] = requester_id
        return False, None, None

    def full_processing_iteration(self):
        for coro_id, result_and_exception in self.results.items():
            result, exception = result_and_exception
            self.register_response(self.called_by[coro_id], result, exception)
            del self.called_by[coro_id]
        
        self.results = type(self.results)()
        self.make_dead()

    def in_work(self) -> bool:
        result: bool = bool(self.results)
        return self.thrifty_in_work(result)

    def _on_coro_del_handler(self, coro: CoroWrapperBase) -> bool:
        if coro.coro_id in self.results:
            raise RuntimeError(f'CoroWrapperBase is already in results list. {coro.coro_id=}, {coro.last_result=}, {coro.exception=}')
        
        self.results[coro.coro_id] = (coro.last_result, coro.exception)
        self.make_live()
        return True


async def arun_coro_fast(i: Interface, coro_worker: AnyWorker, *args, **kwargs) -> Any:
    coro_type: CoroType = find_coro_type(coro_worker)
    if CoroType.awaitable == coro_type:
        return await coro_worker(i, *args, **kwargs)
    else:
        return await i(RunCoro, coro_worker, *args, **kwargs)


def run_coro_fast(i: Interface, coro_worker: AnyWorker, *args, **kwargs) -> Any:
    coro_type: CoroType = find_coro_type(coro_worker)
    if CoroType.greenlet == coro_type:
        return coro_worker(i, *args, **kwargs)
    else:
        return i(RunCoro, coro_worker, *args, **kwargs)


async def arun_coro(coro_worker: AnyWorker, *args, **kwargs) -> Any:
    i: Interface = current_interface()
    coro_type: CoroType = find_coro_type(coro_worker)
    if CoroType.awaitable == coro_type:
        return await coro_worker(i, *args, **kwargs)
    else:
        return await i(RunCoro, coro_worker, *args, **kwargs)


def run_coro(coro_worker: AnyWorker, *args, **kwargs) -> Any:
    i: Interface = current_interface()
    coro_type: CoroType = find_coro_type(coro_worker)
    if CoroType.greenlet == coro_type:
        return coro_worker(i, *args, **kwargs)
    else:
        return i(RunCoro, coro_worker, *args, **kwargs)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['default_env_path_and_params', 'Db', 'DbRequest', 'KeyType', 'RawKeyType', 'ValueType', 'RawValueType', 
           'DbId', 'EnvId', 'DbName', 'DbKeyError', 'EnvInfo']

from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_tools.await_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.asyncio_loop import *
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import CoroPriority
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.timer_func_runner import *
from cengal.parallel_execution.coroutines.coro_standard_services.instance import *
from cengal.file_system.file_manager import path_relative_to_current_dir
from cengal.time_management.cpu_clock_cycles import perf_counter
from cengal.data_manipulation.serialization import *
from cengal.introspection.inspect import get_exception
from cengal.file_system.app_fs_structure.app_dir_path import AppDirPath, AppDirectoryType
from cengal.file_system.path_manager import RelativePath
from cengal.code_flow_control.args_manager import ArgsKwargs
from cengal.text_processing.text_processing import to_identifier
from cengal.math.numbers import RationalNumber
from typing import Hashable, Tuple, List, Any, Dict, Callable, Sequence, NamedTuple, OrderedDict as OrderedDictType
import sys
import os
import asyncio
try:
    import lmdb
except ImportError:
    from warnings import warn
    warn('''WARNING: `lmdb` library is not installed. Db service will not work.
         To install `lmdb` use: `pip install lmdb`''')
    raise

from os.path import normpath
from uuid import uuid4
from collections import OrderedDict, namedtuple
from functools import update_wrapper
import inspect


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


SingleKeyType = Union[bytes, str]
CompoundKeyType = Union[List[SingleKeyType], Set[SingleKeyType], Tuple[SingleKeyType], List['CompoundKeyType'], Set['CompoundKeyType'], Tuple['CompoundKeyType']]
KeyType = Union[SingleKeyType, CompoundKeyType]
NormalizedCompoundKeyType = Union[Tuple[SingleKeyType], Tuple['NormalizedCompoundKeyType']]
NormalizedKeyType = Union[SingleKeyType, NormalizedCompoundKeyType]
RawKeyType = bytes  # By default record keys are limited to 511 bytes in length, however this can be adjusted by rebuilding the library. The compile-time key length can be queried via Environment.max_key_size()
ValueType = Any
RawValueType = bytes
DbId = Hashable
DbName = bytes
EnvId = Hashable


class KeyInfo:
    def __init__(self, key: KeyType, db_id: DbId = None, env_id: EnvId = None):
        self.key: KeyType = key
        self.db_id: DbId = db_id
        self.env_id: EnvId = env_id


ValueInfo = NamedTuple('ValueInfo', [('value', ValueType), ('dupdata', bool), ('overwrite', bool), ('append', bool)])
VI = ValueInfo
RawValueInfo = NamedTuple('RawValueInfo', [('value', RawValueType), ('dupdata', bool), ('overwrite', bool), ('append', bool)])
RVI = RawValueInfo


default_env_path_and_params: ArgsKwargs = ArgsKwargs(
    env_path=None,
    can_be_written_externally=False,
    map_size=20 * 1024**2, 
    writemap=True, 
    max_dbs=10,
    map_async=True, 
    lock=False, 
    metasync=False, 
    sync=False, 
    meminit=False,
)


default_env_params: ArgsKwargs = ArgsKwargs(
    can_be_written_externally=False,
    map_size=20 * 1024**2, 
    writemap=True, 
    max_dbs=10,
    map_async=True, 
    lock=False, 
    metasync=False, 
    sync=False, 
    meminit=False,
)


class EnvInitInfo:
    def __init__(self, env_id: EnvId, env_path: str, can_be_written_externally: bool, *args, **kwargs):
        self.env_id: EnvId = env_id
        self.env_path: str = normpath(env_path)
        self.can_be_written_externally: bool = can_be_written_externally
        self.args: Tuple = (self.env_path,) + args
        self.kwargs: Dict = kwargs


class EnvInfo:
    db_name_prefix = '__db_name_key_16318cf6_3e16_4881_a22f_328aa41c0d4f__'
    db_name_prefix_bytes = b'__db_name_key_16318cf6_3e16_4881_a22f_328aa41c0d4f__'

    def __init__(self, init_info: EnvInitInfo, env: Optional[lmdb.Environment] = None):
        self.init_info: EnvInitInfo = init_info
        if not os.path.exists(init_info.env_path) or (os.path.exists(init_info.env_path) and not os.path.isdir(init_info.env_path)):
            os.makedirs(init_info.env_path)
        
        self.env: lmdb.Environment = lmdb.Environment(*init_info.args, **init_info.kwargs) if env is None else env
        self.env_id: EnvId = init_info.env_id
        self.databases: Dict[DbId, lmdb._Database] = dict()
        self.db_names: Dict[DbId, DbName] = dict()
        self.prepare_main_db()
    
    @staticmethod
    def gen_db_name_from_db_id(db_id: DbId) -> Union[None, bytes]:
        if db_id is None:
            return None
        
        return f'{EnvInfo.db_name_prefix}{db_id}'.encode('utf-8')
    
    def db_name_by_db_id(self, db_id: DbId) -> Union[None, bytes]:
        try:
            return self.db_names[db_id]
        except KeyError:
            raise UnknownEnvDBError(self.env_id, db_id)
    
    def db_by_db_id(self, db_id: DbId) -> lmdb._Database:
        try:
            return self.databases[db_id]
        except KeyError:
            raise UnknownEnvDBError(self.env_id, db_id)
    
    def open_db(self, db_id: DbId, *args, **kwargs) -> lmdb._Database:
            db_name: Union[None, bytes] = self.gen_db_name_from_db_id(db_id)
            new_db: lmdb._Database = self.env.open_db(db_name, *args, **kwargs)
            self.databases[db_id] = new_db
            self.db_names[db_id] = db_name
            return new_db
    
    def prepare_main_db(self) -> lmdb._Database:
        self.open_db(None)
    
    def close(self):
        self.env.close()


class DbKeyError(KeyError):
    def __init__(self, key_info: Tuple[KeyType, DbId], *args: object) -> None:
        super().__init__(*args)
        self.key_info: Tuple[KeyType, DbId] = key_info


class DBError(Exception):
    def __init__(self, db_id: DbId, original_exception: Exception, *args):
        super().__init__(*args)
        self.db_id: DbId = db_id
        self.original_exception = original_exception
    
    @staticmethod
    def from_exception(db_id: DbId) -> 'DBError':
        return DBError(db_id, get_exception())


class UnknownEnvError(Exception):
    pass


class UnknownEnvDBError(Exception):
    pass


class WrongKeyTypeError(TypeError):
    pass


class RawKeyIsTooLargeError(ValueError):
    pass


class DefaultDbEnvironmentCanNotBeClosedManualyError(Exception):
    pass


class NormalizedCompoundKey:
    def __init__(self, normalized_key: NormalizedKeyType) -> None:
        self.normalized_key: NormalizedKeyType = normalized_key
    
    def __call__(self) -> Any:
        return self.normalized_key

    @staticmethod
    def from_key(key: KeyType):
        return NormalizedCompoundKey(normalize_compound_key(key))


NCK = NormalizedCompoundKey
InputKeyType = Union[NormalizedCompoundKey, KeyType]


def is_normalized_compound_key(key: KeyType) -> bool:
    if isinstance(key, (bytes, str, int, float)):
        return True
    elif isinstance(key, tuple):
        return all(is_normalized_compound_key(item) for item in key)
    else:
        return False


def normalize_compound_key(key: KeyType) -> NormalizedKeyType:
    if isinstance(key, NormalizedCompoundKey):
        return key()
    elif is_normalized_compound_key(key):
        return key
    elif isinstance(key, list):
        need_to_sort: bool = False
    elif isinstance(key, (set, frozenset)):
        need_to_sort = True
    elif isinstance(key, dict):
        key = tuple(key.items())
        need_to_sort = True
    else:
        raise WrongKeyTypeError(f'Wrong key type: {type(key)}: {key}')

    new_key = list()
    for item in key:
        new_key.append(normalize_compound_key(item))
    
    if need_to_sort:
        new_key.sort()
    
    key = tuple(new_key)

    return key


class DbRequest(ServiceRequest):
    def __init__(self, env_id: EnvId = None, db_id: DbId = None, needs_sync: bool = False, can_wait: bool = False):
        super().__init__()
        self.env_id: EnvId = env_id
        self.db_id: DbId = db_id
        self.needs_sync: bool = needs_sync
        self.provide_to_request_handler = True
        self.can_wait: bool = can_wait  # TODO: implement. If True then request can wait for a next iteration in attempt to create a bunch of requests. If False then request will be processed immediately.
    
    def _copy(self) -> 'DbRequest':
        return DbRequest(self.env_id, self.db_id, self.needs_sync)
    
    def set_root_path_to_db_environments(self, root_path_to_db_environments: str) -> bool:
        return self._save_to_copy(0, root_path_to_db_environments)
    
    def open_databases(self, db_ids: Set[DbId], *args, **kwargs) -> None:
        return self._save_to_copy(1, db_ids, *args, **kwargs)
    
    def drop_db(self, db_id: DbId, delete: bool = False) -> None:
        return self._save_to_copy(2, db_id, delete)
    
    def sync(self) -> None:
        return self._save_to_copy(3)
    
    def wait_sync(self) -> None:  # TODO: implement
        return self._save_to_copy(22)
    
    def set_sync_timeout(self, timeout: RationalNumber) -> None:  # TODO: implement
        return self._save_to_copy(23, timeout)
    
    def get(self, key: InputKeyType) -> ValueType:
        return self._save_to_copy(4, key)
    
    def get_first(self) -> Tuple[NormalizedKeyType, ValueType]:
        # Returns first item in DB
        return self._save_to_copy(14)
    
    def get_last(self) -> Tuple[NormalizedKeyType, ValueType]:
        # Returns last item in DB
        return self._save_to_copy(15)
    
    def get_items(self, db_keys: Sequence[InputKeyType]) -> OrderedDictType[NormalizedKeyType, Tuple[ValueType, Optional[Exception]]]:
        return self._save_to_copy(5, db_keys)
    
    def get_n_items(self, desired_key: InputKeyType, num: Optional[int] = None) -> OrderedDictType[NormalizedKeyType, Tuple[ValueType, Optional[Exception]]]:
        return self._save_to_copy(16, desired_key, num, reverse=False)
    
    def get_reverse_n_items(self, desired_key: InputKeyType, num: Optional[int] = None) -> OrderedDictType[NormalizedKeyType, Tuple[ValueType, Optional[Exception]]]:
        return self._save_to_copy(16, desired_key, num, reverse=True)
    
    def get_items_range(self, first_desired_key: InputKeyType, last_desired_key: InputKeyType, num: Optional[int] = None) -> OrderedDictType[NormalizedKeyType, Tuple[ValueType, Optional[Exception]]]:
        return self._save_to_copy(17, first_desired_key, last_desired_key, num, reverse=False)
    
    def get_reverse_items_range(self, first_desired_key: InputKeyType, last_desired_key: InputKeyType, num: Optional[int] = None) -> OrderedDictType[NormalizedKeyType, Tuple[ValueType, Optional[Exception]]]:
        return self._save_to_copy(17, first_desired_key, last_desired_key, num, reverse=True)
    
    def get_all_items(self) -> OrderedDictType[NormalizedKeyType, Tuple[ValueType, Optional[Exception]]]:
        # Returns all items from DB
        return self._save_to_copy(6)
    
    def pop(self, key: InputKeyType) -> ValueType:  # TODO: implement
        return self._save_to_copy(24, key)
    
    def pop_items(self, db_keys: Sequence[InputKeyType]) -> OrderedDictType[NormalizedKeyType, Tuple[ValueType, Optional[Exception]]]:  # TODO: implement
        return self._save_to_copy(25, db_keys)
    
    def put(self, key: InputKeyType, value: Optional[Union[ValueType, ValueInfo]] = None) -> RawValueType:
        return self._save_to_copy(7, key, value)
    
    def put_items(self, db_items: Dict[InputKeyType, Union[ValueType, ValueInfo]]) -> Dict[DbId, Dict[RawKeyType, Tuple[RawValueType, Optional[Exception]]]]:
        return self._save_to_copy(8, db_items)
    
    def increment(self, key: InputKeyType, value: RationalNumber = 1) -> RationalNumber:  # TODO: implement
        return self._save_to_copy(19, key, value)
    
    inc = increment
    
    def increment_items(self, db_items: Dict[InputKeyType, Union[RationalNumber, ValueInfo]]) -> OrderedDictType[NormalizedKeyType, Tuple[RationalNumber, Optional[Exception]]]:  # TODO: implement
        return self._save_to_copy(8, db_items)
    
    inc_items = increment_items
    
    def replace(self, key: InputKeyType, value: Optional[Union[ValueType, ValueInfo]] = None) -> Tuple[RawValueType, ValueType]:  # TODO: implement
        return self._save_to_copy(20, key, value)
    
    def replace_items(self, db_items: Dict[InputKeyType, Union[ValueType, ValueInfo]]) -> Dict[DbId, Dict[RawKeyType, Tuple[RawValueType, ValueType, Optional[Exception]]]]:  # TODO: implement
        return self._save_to_copy(21, db_items)
    
    def delete(self, key: InputKeyType) -> RawValueType:  # TODO: finish an implementation
        return self._save_to_copy(9, key)
    
    def delete_kv(self, key: InputKeyType, value: ValueType) -> RawValueType:  # TODO: finish an implementation
        return self._save_to_copy(10, key, value)
    
    def delete_items(self, db_items: Set[InputKeyType]) -> Dict[DbId, Dict[RawKeyType, Tuple[RawValueType, Optional[Exception]]]]:  # TODO: finish an implementation
        return self._save_to_copy(11, db_items)
    
    def delete_kv_items(self, db_items: Dict[InputKeyType, Tuple[ValueType]]) -> Dict[DbId, Dict[RawKeyType, Tuple[RawValueType, Optional[Exception]]]]:  # TODO: finish an implementation
        return self._save_to_copy(12, db_items)
    
    def execute_in_transaction(self, env_id: EnvId, callable_or_coro: Union[Callable, AnyWorker]) -> None:  # TODO: implement
        # Will execute given callable or coroutine in transaction (after all other queues will be processed)
        # It will run in current thread. So, it will block current thread until it will be finished if it is not a coroutine.
        # On the other hand it will lock environment and all databases in it until coroutine will be finished.
        return self._save_to_copy(29, env_id, callable_or_coro)

    def open_db_environment(self, env_id: EnvId, env_path: Union[None, str], can_be_written_externally: bool, *args, **kwargs) -> EnvId:
        return self._save_to_copy(13, env_id, env_path, can_be_written_externally, *args, **kwargs)
    
    def close_db_environment(self, env_id: EnvId) -> None:
        return self._save_to_copy(18, env_id)
    
    def lock_databases(self, db_names: Optional[Set[DbId]] = None) -> None:  # TODO: implement
        # Lock all databases if db_names is None. Databases will be released automatically wnen coroutine execution will be finished
        return self._save_to_copy(26, db_names)
    
    def try_lock_databases(self, db_names: Optional[Set[DbId]] = None) -> bool:  # TODO: implement
        # Tries to lock all databases if db_names is None. Returns True if try was successfull. False otherwise. Databases will be released automatically wnen coroutine execution will be finished
        return self._save_to_copy(27, db_names)
    
    def unlock_databases(self, db_names: Optional[Set[DbId]] = None) -> None:  # TODO: implement
        # Unlock all databases if db_names is None
        return self._save_to_copy(28, db_names)


# class TransactionContextManager(DbRequest):
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self._journal_db_id: DbName = b'__journal__'
#         self._transaction: Optional[lmdb.Transaction] = None
#         self._transaction_context_manager: Optional[lmdb.Transaction] = None
#         self._transaction_history: List[DbRequest] = list()
#         self._committed: bool = False

#     @staticmethod
#     def from_request(request: DbRequest) -> 'TransactionContextManager':
#         return TransactionContextManager(request.env_id, request.db_id)
    
#     def to_request(self) -> DbRequest:
#         return DbRequest(self.env_id, self.db_id)
    
#     @staticmethod
#     def request_to_data(request: DbRequest) -> Dict:
#         result = dict()
#         result['env_id'] = request.env_id
#         result['db_id'] = request.db_id
#         result['needs_sync'] = request.needs_sync
#         result['request_type'] = request.request_type
#         result['args'] = request.args
#         result['kwargs'] = request.kwargs
#         result['provide_to_request_handler'] = request.provide_to_request_handler
#         return result
    
#     @staticmethod
#     def request_from_data(data: Dict) -> DbRequest:
#         request = DbRequest(data['env_id'], data['db_id'], data['needs_sync'])
#         request.request_type = data['request_type']
#         request.args = data['args']
#         request.kwargs = data['kwargs']
#         request.provide_to_request_handler = data['provide_to_request_handler']
#         return request
    
#     async def _commit(self):
#         i: Interface = current_interface()
#         try:
#             await i(DbRequest(self.env_id, self._journal_db_id, True).lock_databases({self._journal_db_id}))
#             last_committed_id: int = await i(DbRequest(self.env_id, self._journal_db_id, True).get(1))
#             last_id: int = await i(DbRequest(self.env_id, self._journal_db_id, True).get(0))
#             our_id = last_id
#             for request in self._transaction_history:
#                 our_id += 1
#                 await i(DbRequest(self.env_id, self._journal_db_id, True).put(our_id, TransactionContextManager.request_to_data(request)))

#             last_id: int = await i(DbRequest(self.env_id, self._journal_db_id, True).replace(0, our_id))
#             for request in self._transaction_history:
#                 await i(request)
#         except:
#             ...
#         finally:
#             await i(DbRequest(self.env_id, self._journal_db_id, True).unlock_databases({self._journal_db_id}))
    
#     def _save_to_copy(self, __request__type__: int, *args, **kwargs) -> ServiceRequest:
#         request: DbRequest = super()._save_to_copy(__request__type__, *args, **kwargs)
#         self._transaction_history.append(request)
#         return request
    
#     def __enter__(self) -> 'TransactionContextManager':
#         self._transaction = self.env.begin(write=True)
#         self._transaction_context_manager = self._transaction.__enter__()
#         return self
    
#     def __exit__(self, exc_type, exc_val, exc_tb) -> None:
#         self._transaction_context_manager.__exit__(exc_type, exc_val, exc_tb)
#         self._transaction = None
#         self._transaction_context_manager = None
    
#     async def __aenter__(self) -> 'TransactionContextManager':
#         i: Interface = current_interface()
#         db_service: Db = i._loop.get_service_instance(Db)
#         if self.env_id not in db_service.db_environments:
#             raise UnknownEnvError(self.env_id)
        
#         env_info: EnvInfo = self.db_environments[self.env_id]
#         if self._journal_db_id not in env_info.databases:
#             await i(self.to_request().open_databases({self._journal_db_id}))

#         return self
    
#     async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
#         if (exc_type is None) and (exc_val is None) and (exc_tb is None):
#             await self._commit()


def check_request(method: Callable):
    def wrapper(self: 'Db', request: DbRequest, *args, **kwargs) -> ServiceProcessingResponse:
        if request.env_id not in self.db_environments:
            return True, None, UnknownEnvError(request.env_id)
        
        if request.db_id not in self.db_environments[request.env_id].databases:
            return True, None, UnknownEnvDBError(request.env_id, request.db_id)
        
        return method(self, request, *args, **kwargs)
    
    original_func_sign: inspect.Signature = inspect.signature(method)
    update_wrapper(wrapper, method)
    wrapper.__signature__ = original_func_sign.replace(parameters=tuple(original_func_sign.parameters.values()), return_annotation=original_func_sign.return_annotation)
    return wrapper
    

class Db(Service, EntityStatsMixin):
    def __init__(self, loop: CoroSchedulerType):
        super(Db, self).__init__(loop)
        self.default_env_name: str = '__default__.dbenv'
        self.default_envs_dir: str = 'db_envs'
        # self.drop_db_requests: Dict[CoroID, Tuple[Hashable, bool]] = dict()
        # self.drop_db_requests: Dict[EnvId, Dict[CoroID, Tuple[DbId, bool]]] = dict()
        self.drop_db_requests: Dict[EnvId, Dict[DbId, Tuple[bool, Set[CoroID]]]] = dict()
        # self.read_queue: List[Tuple[CoroID, Tuple[RawKeyType, DbId]]] = list()
        self.read_queue: Dict[EnvId, Dict[DbId, Dict[RawKeyType, Set[CoroID]]]] = dict()
        self.massive_read_queue: Dict[EnvId, Dict[CoroID, Dict[DbId, List[RawKeyType]]]] = dict()
        # self.massive_read_queue: List[Tuple[CoroID, Set[Tuple[KeyType, DbId]]]] = list()
        self.put_queue: Dict[EnvId, Dict[DbId, Dict[RawKeyType, List[Union[RawValueType, RawValueInfo]]]]] = dict()
        self.data_cache: Dict[EnvId, Dict[DbId, Dict[RawKeyType, List[Union[RawValueType, RawValueInfo]]]]] = dict()
        # self.data_cache: Dict[EnvId, Dict[Tuple[RawKeyType, DbId], RawValueType]] = dict()
        self.max_data_cache_size: Union[None, int] = 10000
        self.deletion_cache: Dict[EnvId, Dict[DbId, Set[RawKeyType]]] = dict()
        self.kv_deletion_cache: Dict[EnvId, Dict[DbId, Dict[RawKeyType, List[RawValueType]]]] = dict()
        # self.kv_deletion_cache: Dict[Tuple[Hashable, Hashable], Any] = dict()
        self.get_first_queue: Dict[EnvId, Dict[DbId, Set[CoroID]]] = dict()
        self.get_last_queue: Dict[EnvId, Dict[DbId, Set[CoroID]]] = dict()
        self.get_n_items_queue: Dict[EnvId, Dict[CoroID, Tuple[DbId, RawKeyType, int, bool]]] = dict()
        self.get_items_range_queue: Dict[EnvId, Dict[CoroID, Tuple[DbId, RawKeyType, RawKeyType, int, bool]]] = dict()
        # self.get_all_items_queue: List[Tuple[CoroID, DbId, EnvId]] = list()
        self.get_all_items_queue: Dict[EnvId, List[Tuple[CoroID, DbId]]] = dict()
        self.open_db_environment_requests: Dict[CoroID, Tuple[EnvId, Union[None, str], bool, Tuple, Dict]] = dict()
        self.root_path_to_db_environments_rel: RelativePath = None
        self.app_name_waiter: CoroWrapperBase = None
        self.default_db_environment: lmdb.Environment = None
        self.db_environments: Dict[EnvId, EnvInfo] = dict()
        # self.databases: Dict[Hashable, Any] = dict()
        # self.db_names: Dict[DbId, DbName] = dict()
        self.async_loop = None
        self.sync_time_interval = 1.0
        self.last_sync_time = perf_counter()
        self.envs_need_to_be_sync: Set[DbId] = set()
        self.envs_in_sync: Set[DbId] = set()
        self.force_sync: Set[EnvId] = set()
        self.sync_an_each_write: bool = False
        self.write_locked: Set[EnvId] = set()
        self.writes_num: int = 0
        self.reads_num: int = 0
        self.deletes_num: int = 0
        self.db_drops_num: int = 0
        self.write_locked_coro_id: Set[CoroID] = set()
        self.wake_up_handle_registered: bool = False
        # self.serializer = best_serializer_for_standard_data((
        #                                     DataFormats.binary,
        #                                     DataFormats.messagepack,
        #                                     Tags.can_use_bytes,
        #                                     Tags.decode_str_as_str,
        #                                     Tags.decode_list_as_list,
        #                                     Tags.decode_bytes_as_bytes,
        #                                     Tags.superficial,
        #                                     Tags.current_platform,
        #                                     Tags.multi_platform,
        #                                 ),
        #                                 TestDataType.small,
        #                                 0.1)
        self.serializer = Serializer(Serializers.msgspec_messagepack)
        self._request_workers = {
            0: self._on_set_root_path_to_db_environments,
            1: self._on_open_databases,
            2: self._on_drop_db,
            3: self._on_sync,
            4: self._on_get,
            5: self._on_get_items,
            6: self._on_get_all_items,
            7: self._on_put,
            8: self._on_put_items,
            9: self._on_delete,
            10: self._on_delete_kv,
            11: self._on_delete_items,
            12: self._on_delete_kv_items,
            13: self._on_open_db_environment,
            14: self._on_get_first,
            15: self._on_get_last,
            16: self._on_get_n_items,
            17: self._on_get_items_range,
            18: self._on_close_db_environment,
        }

    # TODO: sync with last implementation
    def get_entity_stats(self, stats_level: 'EntityStatsMixin.StatsLevel' = EntityStatsMixin.StatsLevel.debug) -> Tuple[str, Dict[str, Any]]:
        return type(self).__name__, {
            'db_env_ids': list(self.db_environments.keys()),
            'writes num': self.writes_num,
            'reads num': self.reads_num,
            'deletes num': self.deletes_num,
            'db drops num': self.db_drops_num,
        }
    
    def norm_key(self, key: InputKeyType) -> NormalizedKeyType:
        return normalize_compound_key(key)

    def raw_key(self, env_or_id: Union[lmdb.Environment, EnvId], key: InputKeyType) -> RawKeyType:
        raw_key: bytes = self.serializer.dumps(normalize_compound_key(key))
        if isinstance(env_or_id, lmdb.Environment):
            env = env_or_id
        else:
            env = self.db_environments[env_or_id].env
        
        if len(raw_key) > env.max_key_size():
            raise RawKeyIsTooLargeError(f'Raw form ({raw_key=}) of the key ({key=}) is too large: {len(raw_key)} > {env.max_key_size()}')

    def destroy(self):
        # TODO: we need to use some loop destroy service in order to put our coro which will write all pending queues,
        # sync envirounments and close them. Also we need to prevent new requests from being processed.
        db_environments_values = self.db_environments.values()
        self.db_environments = type(self.db_environments)()
        self.default_db_environment = None
        for env_info in db_environments_values:
            env_info.close()

    def single_task_registration_or_immediate_processing(
            self, *args, **kwargs) -> ServiceProcessingResponse:
        result = self.try_resolve_request(*args, **kwargs)
        if result is None:
            self._ensure_default_db_environment()
            return True, None, None
        else:
            return result

    def _ensure_default_db_environment(self) -> bool:
        if self.default_db_environment is None:
            if self.root_path_to_db_environments_rel is None:
                if self.app_name_waiter is None:
                    async def coro(i: Interface, self: 'Db'):
                        app_name_for_fs = await i(InstanceRequest().wait('app_name_for_fs'))
                        app_data_dir_path_type: AppDirectoryType = await i(InstanceRequest().wait('app_data_dir_path_type'))
                        app_dir_path: AppDirPath = await i(InstanceRequest().wait(AppDirPath))
                        app_data_dir_path: str = app_dir_path.cached(app_data_dir_path_type, app_name_for_fs)
                        self.root_path_to_db_environments_rel = RelativePath(RelativePath(app_data_dir_path)(self.default_envs_dir))
                        self._init_default_db_env()
                        self.app_name_waiter = None
                        self.make_live()
                    
                    self.app_name_waiter = put_root_from_other_service(self, coro, self)
                    # self.app_name_waiter.is_background_coro = True
                
                self.make_dead()
                return False
            else:
                self._init_default_db_env()
                return True
        else:
            return True

    def full_processing_iteration(self):
        # TODO: combine all queues into one single transaction by env_id. Or at most two transactions: read and write.
        # This will improve performance and will give ability to move transations to other threads or even processess if needed.

        # TODO: since DB can not handle big number of transactions per secons, it is better to combine all requests and process 
        # them at most as once a milisecond (it is frequently engough).

        if not self._ensure_default_db_environment():
            return
        
        if self.force_sync:
            self.envs_need_to_be_sync |= self.force_sync
            self.force_sync = set()
        
        put_queue_buff = self.put_queue
        self.put_queue = type(put_queue_buff)()
        
        data_cache_buff = self.data_cache  # will be cleared at the end of the iteration if necessary
        
        read_queue_buff = self.read_queue
        self.read_queue = type(read_queue_buff)()
        
        massive_read_queue_buff = self.massive_read_queue
        self.massive_read_queue = type(massive_read_queue_buff)()
        
        deletion_cache_buff = self.deletion_cache
        self.deletion_cache = type(deletion_cache_buff)()
        
        kv_deletion_cache_buff = self.kv_deletion_cache
        self.kv_deletion_cache = type(kv_deletion_cache_buff)()
        
        get_all_items_queue_buff = self.get_all_items_queue
        self.get_all_items_queue = type(get_all_items_queue_buff)()

        get_first_queue_buff = self.get_first_queue
        self.get_first_queue = type(get_first_queue_buff)()

        get_last_queue_buff = self.get_last_queue
        self.get_last_queue = type(get_last_queue_buff)()

        get_n_items_queue_buff = self.get_n_items_queue
        self.get_n_items_queue = type(get_n_items_queue_buff)()

        get_items_range_queue_buff = self.get_items_range_queue
        self.get_items_range_queue = type(get_items_range_queue_buff)()

        open_db_environment_requests_buff: Dict[CoroID, Tuple[EnvId, Union[None, str], bool, Tuple, Dict]] = self.open_db_environment_requests
        self.open_db_environment_requests = type(open_db_environment_requests_buff)()

        # open_db_environment
        for coro_id, request_info in open_db_environment_requests_buff.items():
            env_id, env_path, can_be_written_externally, args, kwargs = request_info
            if env_id in self.db_environments:
                self.register_response(coro_id, self.db_environments[env_id])
            else:
                exception = None
                try:
                    result = self._init_db_env(env_id, env_path, can_be_written_externally, *args, **kwargs)
                except:
                    exception = get_exception()

                self.register_response(coro_id, result, exception)

        # put
        def put_handler(env_info: EnvInfo, put_info: Dict[DbId, Dict[RawKeyType, List[Union[RawValueType, RawValueInfo]]]]):
            try:
                with env_info.env.begin(write=True) as txn:
                    for db_id, db_put_info in put_info.items():
                        if db_id in env_info.databases:
                            for raw_key, values in db_put_info.items():
                                for value in values:
                                    if isinstance(value, RawValueInfo):
                                        value, dupdata, overwrite, append = value
                                    else:
                                        dupdata, overwrite, append = True, True, False
                                    
                                    txn.put(raw_key, value, db=env_info.databases[db_id], dupdata=dupdata, overwrite=overwrite, append=append)
                        
                        self.writes_num += len(db_put_info)
            except lmdb.MapFullError:
                raise DBError.from_exception(db_id)
        
        for env_id, put_info in put_queue_buff.items():
            if env_id in self.db_environments:
                self.envs_need_to_be_sync.add(env_id)
                lmdb_reapplier(self.db_environments[env_id], put_handler, put_info)
        
        # TODO: implement replace* methods processing

        # delete
        for env_id, deletion_cache_buff_db_info in deletion_cache_buff.items():
            if env_id in self.db_environments:
                env_info = self.db_environments[env_id]
                self.envs_need_to_be_sync.add(env_id)
                with env_info.env.begin(write=True) as txn:
                    for db_id, del_keys in deletion_cache_buff_db_info.items():
                        if db_id in env_info.databases:
                            for del_raw_key in del_keys:
                                txn.delete(del_raw_key, db=env_info.databases[db_id])
                                self.deletes_num += 1
                                try:
                                    data_cache_buff[env_id][db_id].pop(del_raw_key, None)
                                except KeyError:
                                    pass

        # delete_kv
        for env_id, kv_deletion_cache_buff_db_info in kv_deletion_cache_buff.items():
            if env_id in self.db_environments:
                env_info = self.db_environments[env_id]
                self.envs_need_to_be_sync.add(env_id)
                with env_info.env.begin(write=True) as txn:
                    for db_id, del_keys in kv_deletion_cache_buff_db_info.items():
                        if db_id in env_info.databases:
                            for del_raw_key, del_raw_values in del_keys.items():
                                for del_raw_value in del_raw_values:
                                    txn.delete(del_raw_key, del_raw_value, db=env_info.databases[db_id])
                                    self.deletes_num += 1
                                    try:
                                        raw_values: List[Union[RawValueType, RawValueInfo]] = data_cache_buff[env_id][db_id][del_raw_key]
                                        new_raw_values: List[Union[RawValueType, RawValueInfo]] = list()
                                        for raw_value in raw_values:
                                            raw_value_original = raw_value
                                            if isinstance(raw_value, RawValueInfo):
                                                raw_value, _, _, _ = raw_value
                                            
                                            if raw_value != del_raw_value:
                                                new_raw_values.append(raw_value_original)
                                            
                                        if new_raw_values:
                                            data_cache_buff[env_id][db_id][del_raw_key] = new_raw_values
                                        else:
                                            data_cache_buff[env_id][db_id].pop(del_raw_key, None)
                                    except KeyError:
                                        pass

        # drop
        drop_db_requests_buff = self.drop_db_requests
        self.drop_db_requests = type(drop_db_requests_buff)()
        
        def drop_handler(env_info: EnvInfo, drop_info: Dict[DbId, Tuple[bool, Set[CoroID]]]):
            for db_id, db_drop_info in drop_info.items():
                delete_db, coro_id = db_drop_info
                if db_id in env_info.databases:
                    try:
                        with env_info.env.begin(write=True) as txn:
                            txn.drop(db=env_info.databases[db_id], delete=delete_db)
                            if delete_db:
                                del env_info.databases[db_id]
                                del env_info.db_names[db_id]
                        
                        self.db_drops_num += 1
                    except lmdb.MapFullError:
                        raise DBError.from_exception(db_id)
                    
                    self.register_response(coro_id, None, UnknownEnvError(env_id))
                else:
                    self.register_response(coro_id, None, UnknownEnvDBError(env_id, db_id))
        
        for env_id, drop_info in drop_db_requests_buff.items():
            if env_id in self.db_environments:
                self.envs_need_to_be_sync.add(env_id)
                lmdb_reapplier(self.db_environments[env_id], drop_handler, drop_info)
            else:
                for db_id, db_drop_info in drop_info.items():
                    delete_db, coro_id = db_drop_info
                    self.register_response(coro_id, None, UnknownEnvError(env_id))

        # get
        def get_item(txn, key_info: Tuple[RawKeyType, DbId, EnvId], data_cache_buff: Dict[EnvId, Dict[DbId, Dict[RawKeyType, List[Union[RawValueType, RawValueInfo]]]]]) -> Tuple[ValueType, Optional[Exception]]:
            key, db_id, env_id = key_info
            need_to_get_from_db = True
            try:
                values = data_cache_buff[env_id][db_id][key]
                if values:
                    value = values[0]
                    if isinstance(value, RawValueInfo):
                        value, _, _, _ = value

                    need_to_get_from_db = False
            except KeyError:
                pass
            
            if need_to_get_from_db:
                value = txn.get(key, db=self.db_environments[env_id].databases[db_id])
                self.reads_num += 1
            
            exception = None
            try:
                if value is None:
                    exception = DbKeyError(key_info)
                else:
                    value = self.serializer.loads(value)
            except:
                exception = get_exception()
            
            return value, exception
        
        # _on_get
        for env_id, read_queue_buff_db_info in read_queue_buff.items():
            if env_id in self.db_environments:
                env_info = self.db_environments[env_id]
                for db_id, read_queue_buff_db_key_info in read_queue_buff_db_info.items():
                    if db_id in env_info.databases:
                        with env_info.env.begin() as txn:
                            for raw_key, coro_ids in read_queue_buff_db_key_info.items():
                                value, exception = get_item(txn, (raw_key, db_id, env_id), data_cache_buff)
                                for coro_id in coro_ids:
                                    self.register_response(coro_id, value, exception)
                    else:
                        for raw_key, coro_ids in read_queue_buff_db_key_info.items():
                            for coro_id in coro_ids:
                                self.register_response(coro_id, None, UnknownEnvDBError(env_id, db_id))
            else:
                for db_id, read_queue_buff_db_key_info in read_queue_buff_db_info.items():
                    for raw_key, coro_ids in read_queue_buff_db_key_info.items():
                        for coro_id in coro_ids:
                            self.register_response(coro_id, None, UnknownEnvError(env_id))
        
        # _on_get_items
        results: Dict[CoroID, Dict[DbId, Dict[NormalizedKeyType, Tuple[ValueType, Optional[Exception]]]]] = dict()
        for env_id, massive_read_queue_buff_coro_info in massive_read_queue_buff.items():
            if env_id in self.db_environments:
                env_info = self.db_environments[env_id]
                for coro_id, read_queue_buff_db_info in massive_read_queue_buff_coro_info.items():
                    if coro_id not in results:
                        results[coro_id] = dict()
                    
                    coro_results = results[coro_id]
                    for db_id, raw_keys in read_queue_buff_db_info.items():
                        if db_id not in coro_results:
                            coro_results[db_id] = OrderedDict()
                        
                        coro_db_results = coro_results[db_id]
                        if db_id in env_info.databases:
                            with env_info.env.begin() as txn:
                                for raw_key in raw_keys:
                                    value, exception = get_item(txn, (raw_key, db_id, env_id), data_cache_buff)
                                    coro_db_results[normalize_compound_key(raw_key)] = (value, exception)
                        else:
                            for coro_id in coro_ids:
                                self.register_response(coro_id, None, UnknownEnvDBError(env_id, db_id))
            else:
                for coro_id, read_queue_buff_db_info in massive_read_queue_buff_coro_info.items():
                    for db_id, raw_keys in read_queue_buff_db_info.items():
                        for coro_id in coro_ids:
                            self.register_response(coro_id, None, UnknownEnvError(env_id))
        
        for coro_id, coro_results in results.items():
            if coro_results:
                db_id = tuple(coro_results.keys())[0]
                self.register_response(coro_id, coro_results[db_id], None)
            else:
                self.register_response(coro_id, OrderedDict(), None)

        # get all items
        for env_id, requests_info in get_all_items_queue_buff.items():
            for request_info in requests_info:
                coro_id, db_id = request_info
                if env_id in self.db_environments:
                    env_info = self.db_environments[env_id]
                    env = env_info.env
                    if db_id in env_info.databases:
                        db = env_info.databases[db_id]
                        with env.begin(db=db) as txn:
                            result = list()
                            exception = None
                            try:
                                result = [(normalize_compound_key(self.serializer.loads(k)), self.serializer.loads(v)) for k, v in txn.cursor() if not k.startswith(EnvInfo.db_name_prefix_bytes)]
                                self.reads_num += len(result)
                            except:
                                exception = get_exception()
                            
                            self.register_response(coro_id, result, exception)
                    else:
                        self.register_response(coro_id, None, UnknownEnvDBError(env_id, db_id))
                else:
                    self.register_response(coro_id, None, UnknownEnvError(env_id))
        
        # get_first
        for env_id, get_first_queue_buff_db_info in get_first_queue_buff.items():
            if env_id in self.db_environments:
                env_info = self.db_environments[env_id]
                for db_id, coro_ids in get_first_queue_buff_db_info.items():
                    if db_id in env_info.databases:
                        db = env_info.databases[db_id]
                        with env_info.env.begin(db=db) as txn:
                            result = None
                            exception = None
                            try:
                                cursor: lmdb.Cursor = txn.cursor()
                                if cursor.first():
                                    key = cursor.key()
                                    key_found: bool = True
                                    while key.startswith(EnvInfo.db_name_prefix_bytes):
                                        key_found = cursor.next_nodup()
                                        if not key_found:
                                            break

                                        key = cursor.key()

                                    if key_found:
                                        value = cursor.value()
                                        result = (normalize_compound_key(self.serializer.loads(cursor.key())), self.serializer.loads(cursor.value()))
                                        self.reads_num += 1
                                    else:
                                        exception = KeyError()
                                else:
                                    exception = KeyError()
                            except:
                                exception = get_exception()
                            
                            for coro_id in coro_ids:
                                self.register_response(coro_id, result, exception)
                    else:
                        for coro_id in coro_ids:
                            self.register_response(coro_id, None, UnknownEnvDBError(env_id, db_id))
            else:
                for db_id, coro_ids in get_first_queue_buff_db_info.items():
                    for coro_id in coro_ids:
                        self.register_response(coro_id, None, UnknownEnvError(env_id))
        
        # get_last
        for env_id, get_last_queue_buff_db_info in get_last_queue_buff.items():
            if env_id in self.db_environments:
                env_info = self.db_environments[env_id]
                for db_id, coro_ids in get_last_queue_buff_db_info.items():
                    if db_id in env_info.databases:
                        db = env_info.databases[db_id]
                        with env_info.env.begin(db=db) as txn:
                            result = None
                            exception = None
                            try:
                                cursor: lmdb.Cursor = txn.cursor()
                                if cursor.last():
                                    key: bytes = cursor.key()
                                    key_found: bool = True
                                    key = cursor.key()
                                    while key.startswith(EnvInfo.db_name_prefix_bytes):
                                        key_found = cursor.prev_nodup()
                                        if not key_found:
                                            break

                                        key = cursor.key()

                                    if key_found:
                                        value: bytes = cursor.value()
                                        result = (normalize_compound_key(self.serializer.loads(cursor.key())), self.serializer.loads(cursor.value()))
                                        self.reads_num += 1
                                    else:
                                        exception = KeyError()
                                else:
                                    exception = KeyError()
                            except:
                                exception = get_exception()
                            
                            for coro_id in coro_ids:
                                self.register_response(coro_id, result, exception)
                    else:
                        for coro_id in coro_ids:
                            self.register_response(coro_id, None, UnknownEnvDBError(env_id, db_id))
            else:
                for db_id, coro_ids in get_last_queue_buff_db_info.items():
                    for coro_id in coro_ids:
                        self.register_response(coro_id, None, UnknownEnvError(env_id))
        
        # get_n_items
        for env_id, get_n_items_queue_buff_coro_info in get_n_items_queue_buff.items():
            if env_id in self.db_environments:
                env_info = self.db_environments[env_id]
                for coro_id, read_queue_buff_db_info in get_n_items_queue_buff_coro_info.items():
                    coro_results: List[Tuple[NormalizedKeyType, ValueType]] = list()
                    db_id, first_desired_raw_key, num_items, reverse = read_queue_buff_db_info
                    if db_id in env_info.databases:
                        db = env_info.databases[db_id]
                        with env_info.env.begin(db=db) as txn:
                            exception = None
                            try:
                                cursor: lmdb.Cursor = txn.cursor()
                                if cursor.set_range(first_desired_raw_key):
                                    if reverse:
                                        cursor_iterator = cursor.iterprev(keys=True, values=True)
                                    else:
                                        cursor_iterator = cursor.iternext(keys=True, values=True)
                                    
                                    for raw_key, raw_value in cursor_iterator:
                                        if raw_key.startswith(EnvInfo.db_name_prefix_bytes):
                                            continue

                                        if (num_items is not None) and (num_items <= 0):
                                            break
                                        
                                        coro_results.append((normalize_compound_key(self.serializer.loads(raw_key)), self.serializer.loads(raw_value)))
                                        self.reads_num += 1
                                        if num_items is not None:
                                            num_items -= 1
                                else:
                                    exception = KeyError()
                            except:
                                exception = get_exception()
                            
                            self.register_response(coro_id, coro_results, exception)
                    else:
                        self.register_response(coro_id, None, UnknownEnvDBError(env_id, db_id))
            else:
                for coro_id, read_queue_buff_db_info in get_n_items_queue_buff_coro_info.items():
                    for coro_id in coro_ids:
                        self.register_response(coro_id, None, UnknownEnvError(env_id))
        
        # get_items_range
        for env_id, get_items_range_queue_buff_coro_info in get_items_range_queue_buff.items():
            if env_id in self.db_environments:
                env_info = self.db_environments[env_id]
                for coro_id, read_queue_buff_db_info in get_items_range_queue_buff_coro_info.items():
                    coro_results: List[Tuple[NormalizedKeyType, ValueType]] = list()
                    db_id, first_desired_raw_key, last_desired_raw_key, num_items, reverse = read_queue_buff_db_info
                    if db_id in env_info.databases:
                        db = env_info.databases[db_id]
                        with env_info.env.begin(db=db) as txn:
                            exception = None
                            try:
                                cursor: lmdb.Cursor = txn.cursor()
                                if cursor.set_range(first_desired_raw_key):
                                    if reverse:
                                        cursor_iterator = cursor.iterprev(keys=True, values=True)
                                    else:
                                        cursor_iterator = cursor.iternext(keys=True, values=True)
                                    
                                    for raw_key, raw_value in cursor_iterator:
                                        if raw_key.startswith(EnvInfo.db_name_prefix_bytes):
                                            continue
                                        
                                        if reverse:
                                            if raw_key < last_desired_raw_key:
                                                break
                                        else:
                                            if raw_key > last_desired_raw_key:
                                                break

                                        if (num_items is not None) and (num_items <= 0):
                                            break

                                        coro_results.append((normalize_compound_key(self.serializer.loads(raw_key)), self.serializer.loads(raw_value)))
                                        self.reads_num += 1
                                        if num_items is not None:
                                            num_items -= 1
                                else:
                                    exception = KeyError()
                            except:
                                exception = get_exception()
                            
                            self.register_response(coro_id, coro_results, exception)
                    else:
                        self.register_response(coro_id, None, UnknownEnvDBError(env_id, db_id))
            else:
                for coro_id, read_queue_buff_db_info in get_items_range_queue_buff_coro_info.items():
                    for coro_id in coro_ids:
                        self.register_response(coro_id, None, UnknownEnvError(env_id))
        
        need_to_sync = self.sync_an_each_write

        # periodic sync
        if (perf_counter() - self.last_sync_time) >= self.sync_time_interval:
            for env_id, env_info in self.db_environments.items():
                self.envs_need_to_be_sync.add(env_id)
                need_to_sync = True

        # sync
        if need_to_sync and self.envs_need_to_be_sync:
            envs_need_to_be_sync_bak = self.envs_need_to_be_sync - self.envs_in_sync
            self.envs_need_to_be_sync = set(self.envs_in_sync)
            for env_id in envs_need_to_be_sync_bak:
                self.sync_in_thread_pool(env_id)
            
            self.last_sync_time = perf_counter()
        
        # invalidate data_cache for envs that can be written externally
        for env_id, env_info in self.db_environments.items():
            if env_info.init_info.can_be_written_externally:
                self.data_cache.pop(env_id, None)
        
        # clear too big caches
        for env_id, env_info in self.db_environments.items():
            if env_id in self.data_cache:
                if len(self.data_cache[env_id]) > self.max_data_cache_size:
                    del self.data_cache[env_id]

        self.make_dead()

    def in_work(self) -> bool:
        result: bool = bool(self.default_db_environment is None) \
                            or bool(self.open_db_environment_requests) \
                            or bool(self.get_first_queue) \
                            or bool(self.get_last_queue) \
                            or bool(self.get_n_items_queue) \
                            or bool(self.get_items_range_queue) \
                            or bool(self.read_queue) \
                            or bool(self.massive_read_queue) \
                            or bool(self.get_all_items_queue)\
                            or bool(self.force_sync) \
                            or bool(self.deletion_cache) \
                            or bool(self.kv_deletion_cache) \
                            or bool(self.drop_db_requests) \
                            or bool(self.put_queue) \
                            or bool(self.envs_need_to_be_sync) \
                            or ((perf_counter() - self.last_sync_time) >= self.sync_time_interval)
        return self.thrifty_in_work(result)
    
    def time_left_before_next_event(self) -> Tuple[bool, Optional[Union[int, float]]]:
        time_since_last_sync_time: float = perf_counter() - self.last_sync_time
        if self.sync_time_interval > time_since_last_sync_time:
            return True, self.sync_time_interval - time_since_last_sync_time
        else:
            return True, 0

    def _init_db_env(self, env_id: EnvId, env_path: Union[None, str], can_be_written_externally: bool, *args, **kwargs) -> EnvInfo:
        if env_id in self.db_environments:
            return self.db_environments[env_id]

        if env_path is None:
            if env_id is None:
                env_path = self.root_path_to_db_environments_rel(self.default_env_name)
            else:
                env_path = self.root_path_to_db_environments_rel(to_identifier(f'{env_id}'))
        else:
            if os.path.isabs(env_path):
                env_path = os.path.normpath(env_path)
            else:
                env_path = self.root_path_to_db_environments_rel(env_path)

        env_init_info: EnvInitInfo = EnvInitInfo(
            env_id, 
            env_path, 
            can_be_written_externally, 
            *args, **kwargs
        )
        env_info: EnvInfo = EnvInfo(env_init_info)
        self.db_environments[env_info.env_id] = env_info
        env_info.env.sync(True)
        return env_info

    def _init_default_db_env(self):
        args, kwargs = default_env_path_and_params()
        env_info: EnvInfo = self._init_db_env(None, *args, **kwargs)
        self.default_db_environment = env_info

    def _on_open_db_environment(self, request: DbRequest, env_id: EnvId, env_path: Union[None, str], can_be_written_externally: bool, *args, **kwargs) -> Tuple[bool, EnvInfo, Exception]:
        self.open_db_environment_requests[self.current_caller_coro_info.coro_id] = (env_id, env_path, can_be_written_externally, args, kwargs)
        self.make_live()
        return False, None, None
    
    @check_request
    def _on_close_db_environment(self, request: DbRequest):
        env_id: EnvId = request.env_id
        
        if env_id is None:
            return True, False, DefaultDbEnvironmentCanNotBeClosedManualyError()
        
        if env_id in self.db_environments:
            self.db_environments[env_id].close()
            del self.db_environments[env_id]
            return True, True, None
        else:
            return True, False, UnknownEnvError(env_id)

    def _on_set_root_path_to_db_environments(self, request: DbRequest, root_path_to_db_environments: str) -> ServiceProcessingResponse:
        self.root_path_to_db_environments_rel = RelativePath(root_path_to_db_environments)
        return True, True, None
    
    @check_request
    def _on_sync(self, request: DbRequest) -> ServiceProcessingResponse:
        if self.put_queue:
            self.force_sync.add(request.env_id)
            self.make_live()
        else:
            # self.default_db_environment.sync(True)
            self.sync_in_thread_pool(request.env_id)
        
        return True, None, None
    
    @check_request
    def _on_get(self, request: DbRequest, key: KeyType) -> ServiceProcessingResponse:
        coro_id = self.current_caller_coro_info.coro_id
        key = self.serializer.dumps(normalize_compound_key(key))
        db_id = request.db_id
        env_id = request.env_id
        
        if env_id in self.data_cache:
            env_cache = self.data_cache[env_id]
            if db_id in env_cache:
                db_cache = env_cache[db_id]
                if key in db_cache:
                    values = db_cache[key]
                    if values:
                        value = values[0]
                        if isinstance(value, RawValueInfo):
                            value, _, _, _ = value
                        
                        value = self.serializer.loads(value)
                        return True, value, None
        
        if env_id not in self.read_queue:
            self.read_queue[env_id] = dict()
        
        read_queue_env = self.read_queue[env_id]
        if db_id not in read_queue_env:
            read_queue_env[db_id] = dict()
        
        read_queue_env_db = read_queue_env[db_id]
        if key not in read_queue_env_db:
            read_queue_env_db[key] = set()
        
        read_queue_env_db_key = read_queue_env_db[key]
        read_queue_env_db_key.add(coro_id)
        self.make_live()
        return False, None, None
    
    @check_request
    def _on_get_first(self, request: DbRequest) -> ServiceProcessingResponse:
        coro_id = self.current_caller_coro_info.coro_id
        db_id = request.db_id
        env_id = request.env_id
        
        if env_id not in self.get_first_queue:
            self.get_first_queue[env_id] = dict()
        
        get_first_queue_env = self.get_first_queue[env_id]
        if db_id not in get_first_queue_env:
            get_first_queue_env[db_id] = set()
        
        get_first_queue_env_db = get_first_queue_env[db_id]
        get_first_queue_env_db.add(coro_id)
        
        self.make_live()
        return False, None, None
    
    @check_request
    def _on_get_last(self, request: DbRequest) -> ServiceProcessingResponse:
        coro_id = self.current_caller_coro_info.coro_id
        db_id = request.db_id
        env_id = request.env_id
        
        if env_id not in self.get_last_queue:
            self.get_last_queue[env_id] = dict()
        
        get_last_queue_env = self.get_last_queue[env_id]
        if db_id not in get_last_queue_env:
            get_last_queue_env[db_id] = set()
        
        get_last_queue_env_db = get_last_queue_env[db_id]
        get_last_queue_env_db.add(coro_id)
        
        self.make_live()
        return False, None, None
    
    @check_request
    def _on_get_n_items(self, request: DbRequest, desired_key: InputKeyType, num: Optional[int] = None, reverse: bool = False) -> Tuple[bool, List[Tuple[NormalizedKeyType, ValueType]], Optional[Exception]]:
        coro_id = self.current_caller_coro_info.coro_id
        desired_key = self.serializer.dumps(normalize_compound_key(desired_key))
        db_id = request.db_id
        env_id = request.env_id

        if env_id not in self.get_n_items_queue:
            self.get_n_items_queue[env_id] = dict()
        
        get_n_items_queue_env = self.get_n_items_queue[env_id]
        get_n_items_queue_env[coro_id] = (db_id, desired_key, num, reverse)
        self.make_live()
        return False, None, None
    
    @check_request
    def _on_get_items_range(self, request: DbRequest, first_desired_key: InputKeyType, last_desired_key: InputKeyType, num: Optional[int] = None, reverse: bool = False) -> Tuple[bool, List[Tuple[NormalizedKeyType, ValueType]], Optional[Exception]]:
        coro_id = self.current_caller_coro_info.coro_id
        first_desired_key = self.serializer.dumps(normalize_compound_key(first_desired_key))
        last_desired_key = self.serializer.dumps(normalize_compound_key(last_desired_key))
        db_id = request.db_id
        env_id = request.env_id

        if env_id not in self.get_items_range_queue:
            self.get_items_range_queue[env_id] = dict()
        
        get_items_range_queue_env = self.get_items_range_queue[env_id]
        get_items_range_queue_env[coro_id] = (db_id, first_desired_key, last_desired_key, num, reverse)
        self.make_live()
        return False, None, None
    
    @check_request
    def _on_get_items(self, request: DbRequest, db_keys: Sequence[InputKeyType]) -> Tuple[bool, Dict[DbId, Dict[NormalizedKeyType, Tuple[ValueType, Optional[Exception]]]], Optional[Exception]]:
        coro_id = self.current_caller_coro_info.coro_id
        db_id = request.db_id
        env_id = request.env_id

        if env_id not in self.massive_read_queue:
            self.massive_read_queue[env_id] = dict()
        
        massive_read_queue_env = self.massive_read_queue[env_id]
        if coro_id not in massive_read_queue_env:
            massive_read_queue_env[coro_id] = dict()
        
        massive_read_queue_env_coro = massive_read_queue_env[coro_id]
        if db_id not in massive_read_queue_env_coro:
            massive_read_queue_env_coro[db_id] = list()
        
        massive_read_queue_env_coro_db = massive_read_queue_env_coro[db_id]
        for key in db_keys:
            massive_read_queue_env_coro_db.append(self.serializer.dumps(normalize_compound_key(key)))

        self.make_live()
        return False, None, None
    
    @check_request
    def _on_get_all_items(self, request: DbRequest) -> Tuple[bool, List[Tuple[NormalizedKeyType, ValueType]], Optional[Exception]]:
        env_id = request.env_id
        db_id = request.db_id
        coro_id = self.current_caller_coro_info.coro_id
        if env_id not in self.get_all_items_queue:
            self.get_all_items_queue[env_id] = list()
        
        self.get_all_items_queue[env_id].append((coro_id, db_id))
        self.make_live()
        return False, None, None
    
    @check_request
    def _on_put(self, request: DbRequest, key: KeyType, value: Union[ValueType, ValueInfo]) -> Tuple[bool, RawValueType, Optional[Exception]]:
        key = self.serializer.dumps(normalize_compound_key(key))
        db_id = request.db_id
        env_id = request.env_id
        
        exception = None
        result = None
        try:
            # self.put_queue
            if env_id not in self.put_queue:
                self.put_queue[env_id] = dict()
            
            env_put_queue = self.put_queue[env_id]
            if db_id not in env_put_queue:
                env_put_queue[db_id] = dict()
            
            env_put_queue_db = env_put_queue[db_id]
            if key not in env_put_queue_db:
                env_put_queue_db[key] = list()
            
            # self.data_cache
            if env_id not in self.data_cache:
                self.data_cache[env_id] = dict()
            
            env_data_cache = self.data_cache[env_id]
            if db_id not in env_data_cache:
                env_data_cache[db_id] = dict()
            
            env_data_cache_db = env_data_cache[db_id]
            if key not in env_data_cache_db:
                env_data_cache_db[key] = list()
            
            # both
            if isinstance(value, ValueInfo):
                result = RawValueInfo(self.serializer.dumps(value.value), value.dupdata, value.overwrite, value.append)
                if env_put_queue_db[key] and not value.dupdata:
                    env_data_cache_db[key][0] = env_put_queue_db[key][0] = result
                else:
                    env_put_queue_db[key].append(result)
                    env_data_cache_db[key].append(result)
            else:
                result = self.serializer.dumps(value)
                if env_put_queue_db[key]:
                    env_data_cache_db[key][0] = env_put_queue_db[key][0] = result
                else:
                    env_put_queue_db[key].append(result)
                    env_data_cache_db[key].append(result)
        except:
            exception = get_exception()
        
        self.make_live()
        return True, result, exception
    
    @check_request
    def _on_put_items(self, request: DbRequest, db_items: Dict[InputKeyType, Union[ValueType, ValueInfo]]) -> Tuple[bool, Dict[DbId, Dict[RawKeyType, Tuple[RawValueType, Optional[Exception]]]], Optional[Exception]]:
        result_items: Dict[DbId, Dict[RawKeyType, Tuple[RawValueType, Optional[Exception]]]] = dict()
        db_id = request.db_id
        env_id = request.env_id

        # self.put_queue
        if env_id not in self.put_queue:
            self.put_queue[env_id] = dict()
        
        env_put_queue = self.put_queue[env_id]
        if db_id not in result_items:
            result_items[db_id] = dict()
        
        result_db_items = result_items[db_id]
        
        if db_id not in env_put_queue:
            env_put_queue[db_id] = dict()
        
        env_put_queue_db = env_put_queue[db_id]

        # self.data_cache
        if env_id not in self.data_cache:
            self.data_cache[env_id] = dict()
        
        env_data_cache = self.data_cache[env_id]
        if db_id not in result_items:
            result_items[db_id] = dict()
        
        result_db_items = result_items[db_id]
        
        if db_id not in env_data_cache:
            env_data_cache[db_id] = dict()
        
        env_data_cache_db = env_data_cache[db_id]

        # both
        for key, value in db_items.items():
            key = self.serializer.dumps(normalize_compound_key(key))
            
            exception = None
            result = None
            try:
                if key not in env_put_queue_db:
                    env_put_queue_db[key] = list()
                
                if key not in env_data_cache_db:
                    env_data_cache_db[key] = list()
                
                if isinstance(value, ValueInfo):
                    result = env_data_cache_db[key][0] = env_put_queue_db[key][0] = RawValueInfo(self.serializer.dumps(value.value), value.dupdata, value.overwrite, value.append)
                    if env_data_cache_db[key] and not value.dupdata:
                        env_data_cache_db[key][0] = env_put_queue_db[key][0] = result
                    else:
                        env_put_queue_db[key].append(result)
                        env_data_cache_db[key].append(result)
                else:
                    result = self.serializer.dumps(value)
                    if env_data_cache_db[key]:
                        env_data_cache_db[key][0] = env_put_queue_db[key][0] = result
                    else:
                        env_put_queue_db[key].append(result)
                        env_data_cache_db[key].append(result)
            except:
                exception = get_exception()
            
            result_db_items[key] = (result, exception)
        
        self.make_live()
        return True, result_items, None
    
    @check_request
    def _on_delete(self, request: DbRequest, key: KeyType) -> Tuple[bool, None, Optional[Exception]]:
        key = self.serializer.dumps(normalize_compound_key(key))
        if key.startswith(EnvInfo.db_name_prefix_bytes):
            return True, None, KeyError(f'Can not delete special key (db info key): {key}')
        
        db_id = request.db_id
        env_id = request.env_id
        
        exception = None
        result = None
        try:
            if env_id not in self.deletion_cache:
                self.deletion_cache[env_id] = dict()
            
            env_deletion_cache = self.deletion_cache[env_id]
            if db_id not in env_deletion_cache:
                env_deletion_cache[db_id] = set()
            
            env_deletion_cache_db = env_deletion_cache[db_id]
            env_deletion_cache_db.add(key)
        except:
            exception = get_exception()
        
        self.make_live()
        return True, result, exception
    
    @check_request
    def _on_delete_kv(self, request: DbRequest, key: InputKeyType, value: ValueType) -> Tuple[bool, RawValueType, Optional[Exception]]:
        raw_key: bytes = self.serializer.dumps(normalize_compound_key(key))
        if raw_key.startswith(EnvInfo.db_name_prefix_bytes):
            return True, None, KeyError(f'Can not delete special raw_key (db info raw_key): {raw_key}')
        
        db_id = request.db_id
        env_id = request.env_id
        
        exception = None
        raw_value = None
        try:
            if env_id not in self.kv_deletion_cache:
                self.kv_deletion_cache[env_id] = dict()
            
            env_kv_deletion_cache = self.kv_deletion_cache[env_id]
            if db_id not in env_kv_deletion_cache:
                env_kv_deletion_cache[db_id] = dict()
            
            env_kv_deletion_cache_db = env_kv_deletion_cache[db_id]
            if raw_key not in env_kv_deletion_cache_db:
                env_kv_deletion_cache_db[raw_key] = list()
            
            raw_value = self.serializer.dumps(value)
            env_kv_deletion_cache_db[raw_key].append(raw_value)
        except:
            exception = get_exception()
        
        self.make_live()
        return True, (raw_key, raw_value), exception
    
    @check_request
    def _on_delete_items(self, request: DbRequest, db_items: Set[InputKeyType]) -> Tuple[bool, Dict[DbId, Dict[RawKeyType, Optional[Exception]]], Optional[Exception]]:
        result_items: Dict[InputKeyType, Tuple[RawKeyType, Optional[Exception]]] = dict()
        db_id = request.db_id
        env_id = request.env_id

        if env_id not in self.deletion_cache:
            self.deletion_cache[env_id] = dict()
        
        env_deletion_cache = self.deletion_cache[env_id]
        if db_id not in env_deletion_cache:
            env_deletion_cache[db_id] = set()
        
        env_deletion_cache_db = env_deletion_cache[db_id]
        for key in db_items:
            exception = None
            try:
                raw_key: bytes = self.serializer.dumps(normalize_compound_key(key))
                if raw_key.startswith(EnvInfo.db_name_prefix_bytes):
                    exception = KeyError(f'Can not delete special key (db info key): {raw_key}')
                else:
                    env_deletion_cache_db.add(raw_key)
            except:
                exception = get_exception()
            
            result_items[key] = (raw_key, exception)

        self.make_live()
        return True, result_items, None
    
    @check_request
    def _on_delete_kv_items(self, request: DbRequest, db_items: Dict[InputKeyType, Tuple[ValueType]]) -> Tuple[bool, Dict[DbId, Dict[RawKeyType, Tuple[RawValueType, Optional[Exception]]]], Optional[Exception]]:
        result_items: Dict[InputKeyType, Tuple[RawKeyType, RawValueType, Optional[Exception]]] = dict()
        db_id = request.db_id
        env_id = request.env_id

        if env_id not in self.kv_deletion_cache:
            self.kv_deletion_cache[env_id] = dict()
        
        env_kv_deletion_cache = self.kv_deletion_cache[env_id]
        if db_id not in env_kv_deletion_cache:
            env_kv_deletion_cache[db_id] = dict()
        
        env_kv_deletion_cache_db = env_kv_deletion_cache[db_id]
        for key, value in db_items.items():
            exception = None
            raw_key = None
            raw_value = None
            try:
                raw_key: bytes = self.serializer.dumps(normalize_compound_key(key))
                if raw_key.startswith(EnvInfo.db_name_prefix_bytes):
                    exception = KeyError(f'Can not delete special key (db info key): {raw_key}')
                else:
                    if raw_key not in env_kv_deletion_cache_db:
                        env_kv_deletion_cache_db[raw_key] = list()
                    
                    raw_value = self.serializer.dumps(value)
                    env_kv_deletion_cache_db[raw_key].append(self.serializer.dumps(value))
            except:
                exception = get_exception()
            
            result_items[key] = (raw_key, raw_value, exception)
        
        self.make_live()
        return True, result_items, None
    
    def _on_open_databases(self, request: DbRequest, db_ids: Set[DbId], *args, **kwargs) -> ServiceProcessingResponse:
        try:
            env_info: EnvInfo = self.db_environments[request.env_id]
        except KeyError:
            exception = UnknownEnvError(request.env_id)
            return True, None, exception
        
        for db_id in db_ids:
            env_info.open_db(db_id, *args, **kwargs)
        
        env_info.env.sync(True)
        return True, None, None
    
    @check_request
    def _on_drop_db(self, request: DbRequest, delete: bool = False) -> ServiceProcessingResponse:
        coro_id = self.current_caller_coro_info.coro_id
        db_id = request.db_id
        env_id = request.env_id

        if env_id not in self.drop_db_requests:
            self.drop_db_requests[env_id] = dict()
        
        drop_db_requests_env = self.drop_db_requests[env_id]

        if db_id not in drop_db_requests_env:
            drop_db_requests_env[db_id] = (False, set())
        
        drop_db_requests_env_db: Tuple[bool, Set[CoroID]] = drop_db_requests_env[db_id]

        if delete and (not drop_db_requests_env_db[0]):
            drop_db_requests_env_db = (True, drop_db_requests_env_db[1])
        
        drop_db_requests_env_db[1].add(coro_id)
        self.make_live()
        return False, None, None
    
    def sync_in_thread_pool(self, env_id: EnvId = None):
        async def sync_db_coro(i: Interface, self: 'Db', env_id: EnvId, asyncio_loop, need_to_ensure_asyncio_loop: bool):
            if need_to_ensure_asyncio_loop:
                asyncio_loop = await i(AsyncioLoop, AsyncioLoopRequest().ensure_loop(None,CoroPriority.low, True))
            else:
                if asyncio_loop is None:
                    asyncio_loop = await i(AsyncioLoop, AsyncioLoopRequest().get())
            
            async def sync_db(self: 'Db', asyncio_loop, env: lmdb.Environment):
                def sync_worker():
                    env.sync(True)
                    self.write_locked.discard(env_id)
                
                await task_in_thread_pool(asyncio_loop, sync_worker)

            env: lmdb.Environment = self.db_environments[env_id].env
            await i(AsyncioLoop, AsyncioLoopRequest().wait(sync_db(self, asyncio_loop, env)))
            self.write_locked_coro_id.discard(i.coro_id)
            self.envs_in_sync.discard(env_id)
            def make_service_live_for_a_next_sync(self: 'Db'):
                self.make_live()
                self.wake_up_handle_registered = False
            
            if not self.wake_up_handle_registered:
                self.wake_up_handle_registered = True
                await i(TimerFuncRunner, self.sync_time_interval, make_service_live_for_a_next_sync, self)

        asyncio_loop = None
        need_to_ensure_asyncio_loop = False
        try:
            asyncio_loop = self._loop.get_service_instance(AsyncioLoop).inline_get()
        except AsyncioLoopWasNotSetError:
            need_to_ensure_asyncio_loop = True

        coro: CoroWrapperBase = put_root_from_other_service(self, sync_db_coro, self, env_id, asyncio_loop, need_to_ensure_asyncio_loop)
        coro.is_background_coro = True
        self.write_locked.add(env_id)
        self.write_locked_coro_id.add(coro.coro_id)
        self.envs_in_sync.add(env_id)


DbRequest.default_service_type = Db


def lmdb_reapplier(env_info: EnvInfo, handler: Callable, *args, **kwargs):
    environment: lmdb.Environment = env_info.env
    failed = True
    while failed:
        need_to_resize: bool = False
        try:
            handler(env_info, *args, **kwargs)
            failed = False
        except DBError as err:
            if isinstance(err.original_exception, lmdb.MapFullError):
                need_to_resize = True
        
        if need_to_resize:
            environment.set_mapsize(environment.info()['map_size'] + 2 * 1024**2)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.code_flow_control.multiinterface_essence import *


class Point:
    def __init__(self, x: float, y: float):
        self.x: float = x
        self.y: float = y


class RectangleModel(EssenceModel):
    def __init__(self, dimension: Point):
        super(RectangleModel, self).__init__()
        self.dimension: Point = dimension


class RectangleInterface(EssenceInterface[RectangleModel]):
    essence_model_class: Type[RectangleModel] = RectangleModel

    def __init__(self, essence_model: EssenceModel):
        super(RectangleInterface, self).__init__(essence_model)

    @em_changer
    def set_dimension(self, dimension: Point) -> NoReturn:
        self.essence_model.dimension = dimension

    def dimension(self) -> Point:
        return self.essence_model.dimension


class StretchingRectangleInterface(RectangleInterface):
    def __init__(self, essence_model: EssenceModel):
        super(StretchingRectangleInterface, self).__init__(essence_model)

    @em_changer
    def stretch_x(self, multiplier: float) -> NoReturn:
        self.essence_model.dimension.x *= multiplier

    @em_changer
    def stretch_y(self, multiplier: float) -> NoReturn:
        self.essence_model.dimension.y *= multiplier


class StretchingSquareInterface(RectangleInterface):
    def __init__(self, essence_model: EssenceModel):
        super(StretchingSquareInterface, self).__init__(essence_model)

    def applicable(self) -> bool:
        return self.essence_model.dimension.x == self.essence_model.dimension.y

    @em_changer
    def stretch(self, multiplier: float) -> NoReturn:
        self.essence_model.dimension.x *= multiplier
        self.essence_model.dimension.y *= multiplier


def rectangle_factory(dimension: Point) -> RectangleModel:
    return simple_em_factory(
            EntityArgsHolder(RectangleModel, dimension),
            [
                RectangleInterface,
                StretchingRectangleInterface,
                StretchingSquareInterface,
            ]
        )
    # result: RectangleModel = RectangleModel(dimension)
    # result.em_add_interface(RectangleInterface)
    # result.em_add_interface(StretchingRectangleInterface)
    # result.em_add_interface(StretchingSquareInterface)
    # return result
    



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['ServerSideClient', 'create_server', 'start_server', 'ServerSideServer', 'CanNotEstablishNamedConnectionError', 'ClientSideClient', 'ClientSideServer']


from typing import Type
from cengal.parallel_execution.asyncio.efficient_streams import *
from cengal.io.named_connections.named_connections_manager import *
from cengal.parallel_execution.asyncio.atasks import create_task
from cengal.code_flow_control.smart_values import ValueCache, ValueExistence
from asyncio.tasks import sleep, Task
from asyncio.events import AbstractServer


class ServerSideClient(NamedConnectionsClient):
    def __init__(self, reader: StreamReaderAbstract, writer: StreamWriterAbstract, server_id: bytes, external_data_full_size: ValueExistence) -> None:
        super().__init__(server_id, external_data_full_size)
        self.reader: StreamReaderAbstract = reader
        self.writer: StreamWriterAbstract = writer
    
    async def serve(self):
        """client_connected_cb

        Returns:
            Any: _description_
        """        
        raise NotImplementedError
    
    # async def server(reader, writer):
    #     reader: StreamReaderAbstract = reader
    #     writer: StreamWriterAbstract = writer

    #     data = await reader.read(100)
    #     message = data.decode()
    #     addr = writer.get_extra_info('peername')

    #     print(f"Received {message!r} from {addr!r}")

    #     print(f"Send: {message!r}")
    #     # for i in range(100):
    #     #     writable_data = data * 1000000
    #     #     # print(f'writable_data len: {len(writable_data)}')
    #     #     writer.write(writable_data)

    #     data_chunk_len = int(cpu_info().l2_cache_size_per_virtual_core / len(data))
    #     writer.start_aw()
    #     stime = perf_counter()
    #     dtime = 0
    #     return_time = 10
    #     index = 0
    #     while dtime < return_time:
    #         index += 1
    #         if 10 <= index:
    #             writer.owrite(pickle.dumps(PickleEncodableClass()))
    #             index = 0
    #         else:
    #             writer.owrite(marshal.dumps(data * randomized_data_size(data_chunk_len)))

    #         await writer.aw_drain_enough()
    #         dtime = perf_counter() - stime
        
    #     await writer.full_drain()

    #     print("Close the connection")
    #     writer.close()


async def create_server(server_side_client_type: Type[ServerSideClient], named_connections_manager: NamedConnectionsManager, stream_manager: StreamManagerAbstract, *args, **kwargs):
    async def client_connected_cb_wrapper(reader: StreamReaderAbstract, writer: StreamWriterAbstract):
        if not stream_manager.try_establish_message_protocol_server_side(reader, writer):
            raise CanNotEstablishNamedConnectionError
        
        server_id: bytes = await reader.read_message()
        named_connection_client: ServerSideClient = server_side_client_type(reader, writer, server_id, named_connections_manager.external_data_full_size)
        named_connections_manager.register_client(named_connection_client)
        await named_connection_client.serve()

    server: AbstractServer = stream_manager.start_server(client_connected_cb_wrapper, *args, **kwargs)
    return server


async def start_server(server: AbstractServer):
    async with server:
        await server.serve_forever()


ServerSideServer = NamedConnectionsServer


# ======================================================


class CanNotEstablishNamedConnectionError(Exception):
    pass


ClientSideClient = NamedConnectionsClient


class ClientSideServer(NamedConnectionsServer):
    def __init__(self, server_id: bytes, named_connections_manager: 'NamedConnectionsManager', stream_manager: StreamManagerAbstract, need_to_full_drain_on_connect: bool, *connection_args, **connection_kwargs) -> None:
        super().__init__(server_id, named_connections_manager)
        self._connected: bool = False
        self._stream_manager: StreamManagerAbstract = stream_manager
        self._need_to_full_drain_on_connect: bool = need_to_full_drain_on_connect
        self._connection_args = connection_args
        self._connection_kwargs = connection_kwargs
        self.autonomous_serving_future: Task = None
        self.reader: StreamReaderAbstract = None
        self.writer: StreamWriterAbstract = None
    
    async def serve(self):
        """Serve

        Returns:
            Any: _description_
        """        
        raise NotImplementedError
    
    async def connect(self):
        if self._connected:
            return
        
        reader, writer = self._stream_manager.open_connection(*self._connection_args, **self._connection_kwargs)
        self.reader = reader
        self.writer = writer
        if not self._stream_manager.try_establish_message_protocol_client_side(reader, writer):
            raise CanNotEstablishNamedConnectionError
        
        writer.optimized_write_message(self.id)
        if self._need_to_full_drain_on_connect:
            await writer.full_drain()
        
        self.autonomous_serving_future = create_task(self.serve)
        self._connected = True
    
    async def create_client(self, client_side_client_type: Type[ClientSideClient]):
        if not self._connected:
            await self.connect()
        
        return self.named_connections_manager().create_client(client_side_client_type, self.id)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.io.used_ports import *
port = purify_ports(used_ports().port(Protocol.tcp, used_ports().tables[Table.user], {PortStatus.no})())
print(f'port: {port}')
ports_iter = used_ports()(Protocol.tcp, used_ports().tables[Table.user], {PortStatus.no, PortStatus.reserved}, 6)
start, stop = purify_ports(ports_iter())
print(f'first port: {start}, last port: {stop}')
start, stop = purify_ports(ports_iter.shift(100))
print(f'first port: {start}, last port: {stop}')
start, stop = purify_ports(ports_iter.shift(35))
print(f'first port: {start}, last port: {stop}')

print('System ports:')
index = 10
for ports_range in used_ports().port(Protocol.tcp, used_ports().tables[Table.system], {PortStatus.no}):
    port = purify_ports(ports_range)
    print(f'\tfree port: {port}')
    index -= 1
    if index <= 0:
        break

print('User ports ranges:')
index = 10
for ports_range in used_ports()(Protocol.tcp, used_ports().tables[Table.user], {PortStatus.no, PortStatus.reserved}, 50):
    start, stop = purify_ports(ports_range)
    print(f'\tfirst port: {start}, last port: {stop}')
    index -= 1
    if index <= 0:
        break

print('Not overlapping User ports ranges:')
index = 10
ports_iter = used_ports()(Protocol.tcp, used_ports().tables[Table.user], {PortStatus.no, PortStatus.reserved}, 50)
for ports_range in ports_iter:
    start, stop = purify_ports(ports_range)
    print(f'\tfirst port: {start}, last port: {stop}')
    ports_iter.put_only(stop + 1)
    index -= 1
    if index <= 0:
        break

print('Not overlapping User ports ranges:')
index = 10
ports_iter = used_ports()(Protocol.tcp, used_ports().tables[Table.ephemeral], {PortStatus.no, PortStatus.reserved, PortStatus.unofficial}, 50)
for ports_range in ports_iter:
    start, stop = purify_ports(ports_range)
    print(f'\tfirst port: {start}, last port: {stop}')
    ports_iter.put_only(stop + 1)
    index -= 1
    if index <= 0:
        break



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import unittest
from cengal.unittest.behavior_stabilizer import UnittestTestCaseState, UnittestTestCaseWithState
from cengal.math.numbers import ae
from cengal.time_management.run_time import RT
from cengal.introspection.inspect import get_exception, pifrl

from cengal.parallel_execution.coroutines.coro_tools.run_in_loop import run_in_loop
from cengal.parallel_execution.coroutines.coro_scheduler import CoroScheduler, CoroSchedulerType, Interface, CoroWrapperBase, CoroID, GreenletExit, AnyWorker, current_interface
from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
from cengal.parallel_execution.coroutines.coro_standard_services.cpu_tick_count_per_second import CpuTickCountPerSecond
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro, PutCoroRequest
from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import RunCoro
from cengal.parallel_execution.coroutines.coro_standard_services.kill_coro import KillCoro
from cengal.parallel_execution.coroutines.coro_standard_services.throw_coro import ThrowCoro
from cengal.parallel_execution.coroutines.coro_standard_services.shutdown_loop import ShutdownLoop
from cengal.parallel_execution.coroutines.coro_standard_services.asyncio_loop import AsyncioLoopRequest
from cengal.parallel_execution.coroutines.coro_standard_services.instance import InstanceRequest
from cengal.parallel_execution.coroutines.coro_standard_services.shutdown_on_keyboard_interrupt import ShutdownOnKeyboardInterrupt
from cengal.parallel_execution.coroutines.coro_standard_services.db.versions.v_1 import *
from cengal.time_management.cpu_clock_cycles import perf_counter
from contextlib import contextmanager, asynccontextmanager
from time import perf_counter
from typing import Tuple, Optional, OrderedDict as OrderedDictType
from shutil import rmtree


async def wait_for_db_env_unlocked(db_request: DbRequest):
    i: Interface = current_interface()
    loop: CoroSchedulerType = i._loop
    db_service: Db = loop.get_service_instance(Db)
    while db_request.env_id in db_service.write_locked:
        await i(Sleep, 0.01)


async def destroy_db_service(with_data: bool = False):
    i: Interface = current_interface()
    loop: CoroSchedulerType = i._loop
    db_service: Db = loop.get_service_instance(Db)
    dir_to_delete = db_service.root_path_to_db_environments_rel('')
    while db_service.write_locked:
        await i(Sleep, 0.01)
    
    loop.unregister_service(Db)
    if with_data:
        rmtree(dir_to_delete)


@asynccontextmanager
async def db(db_request: Optional[DbRequest] = None, destroy_with_data: bool = True):
    i: Interface = current_interface()
    db_service: Db = i._loop.get_service_instance(Db)
    await i(Db)
    while db_service.default_db_environment is None:
        await i(Sleep, 0.01)
    
    if db_request is not None:
        if db_request.env_id is not None:
            args, kwargs = default_env_path_and_params()
            await i(db_request.open_db_environment(db_request.env_id, *args, **kwargs))
        
        await i(db_request.open_databases({db_request.db_id}))

    try:
        yield db_service
    finally:
        db_service: Db = i._loop.get_service_instance(Db)
        while db_service.write_locked:
            await i(Sleep, 0.01)

        await i(db_request.sync())
        await destroy_db_service(destroy_with_data)


async def main(i: Interface, test: AnyWorker, *test_args, **test_kwargs) -> None:
    await i(AsyncioLoopRequest().ensure_loop(interrupt_when_no_requests=True))
    await i(AsyncioLoopRequest().turn_on_loops_intercommunication())
    await i(ShutdownOnKeyboardInterrupt)
    await i(InstanceRequest().set('app_name_for_fs', 'test__db'))
    testing_envs_and_dbs = [
        (None, None),
        (None, 'db_1'),
        ('env_1', None),
        ('env_1', 'db_1'),
    ]
    for env_id, db_id in testing_envs_and_dbs:
        print()
        print(f'env_id: {env_id}, db_id: {db_id}')
        db_request: DbRequest = DbRequest(env_id, db_id)
        await i(RunCoro, test, db_request, *test_args, **test_kwargs)


class TestCaseForDb(unittest.TestCase):
    # @unittest.skip('Skip')
    def test_00_register_service(self):
        pifrl()
        async def coro(i: Interface, db_request: DbRequest) -> None:
            loop: CoroSchedulerType = i._loop
            loop.register_service(Db)
            await i(Sleep, 0.01)
            db_service = loop.get_service_instance(Db)
            print(repr(db_service))

        with RT() as rt:
            run_in_loop(main, coro)
        
        print(f'Run time: {rt.rt}')

    # @unittest.skip('Skip')
    def test_01_init(self):
        pifrl()
        async def coro(i: Interface, db_request: DbRequest) -> None:
            await i(Db)
            await i(Sleep, 0.1)
            loop: CoroSchedulerType = i._loop
            db_service: Db = loop.get_service_instance(Db)
            await i(Db)
            while db_service.default_db_environment is None:
                await i(Sleep, 0.01)
            
            await destroy_db_service(True)

        with RT() as rt:
            run_in_loop(main, coro)
        
        print(f'Run time: {rt.rt}')

    # @unittest.skip('Skip')
    def test_02_put(self):
        pifrl()
        async def coro(i: Interface, db_request: DbRequest) -> None:
            async with db(db_request, False) as db_service:
                await i(db_request.put('key_1', 'value_1'))
                result = await i(db_request.get('key_1'))
                self.assertEqual(result, 'value_1', 'Result is wrong')
                await i(db_request.sync())
                await wait_for_db_env_unlocked(db_request)
                result = await i(db_request.get('key_1'))
                self.assertEqual(result, 'value_1', 'Result is wrong')

        with RT() as rt:
            run_in_loop(main, coro)
        
        print(f'Run time: {rt.rt}')

    # @unittest.skip('Skip')
    def test_03_get(self):
        pifrl()
        async def coro(i: Interface, db_request: DbRequest) -> None:
            result = None
            async with db(db_request, False) as db_service:
                result = await i(db_request.get('key_1'))
            
            print(f'Result: {result}')
            self.assertEqual(result, 'value_1', 'Result is wrong')
            return result

        with RT() as rt:
            run_in_loop(main, coro)
        
        print(f'Run time: {rt.rt}')

    # @unittest.skip('Skip')
    def test_04_cleanup(self):
        pifrl()
        async def coro(i: Interface, db_request: DbRequest) -> None:
            result = None
            async with db(db_request) as db_service:
                pass
            
            return result

        with RT() as rt:
            run_in_loop(main, coro)
        
        print(f'Run time: {rt.rt}')

    # @unittest.skip('Skip')
    def test_05_put_items(self):
        pifrl()
        async def coro(i: Interface, db_request: DbRequest) -> None:
            async with db(db_request, False) as db_service:
                await i(db_request.put_items({'key_1': 'value_1', 'key_2': 'value_2', 'key_3': 'value_3'}))
                result: OrderedDictType = await i(db_request.get_items(['key_1', 'key_2', 'key_3']))
                self.assertEqual([item[0] for item in result.values()], ['value_1', 'value_2', 'value_3'], 'Result is wrong')
                await i(db_request.sync())
                await wait_for_db_env_unlocked(db_request)
                result = await i(db_request.get_items(['key_1', 'key_2', 'key_3']))
                self.assertEqual([item[0] for item in result.values()], ['value_1', 'value_2', 'value_3'], 'Result is wrong')

        with RT() as rt:
            run_in_loop(main, coro)
        
        print(f'Run time: {rt.rt}')

    # @unittest.skip('Skip')
    def test_06_get_items(self):
        pifrl()
        async def coro(i: Interface, db_request: DbRequest) -> None:
            result__get_first = None
            result__get_last = None
            result__get_items = None
            result__get_n_items = None
            result__get_reverse_n_items = None
            result__get_items_range = None
            result__get_reverse_items_range = None
            result__get_all_items = None
            async with db(db_request, False) as db_service:
                result__get_first = await i(db_request.get_first())
                result__get_last = await i(db_request.get_last())
                result__get_items = await i(db_request.get_items(['key_1', 'key_2', 'key_3']))
                result__get_n_items = await i(db_request.get_n_items('key_1', 2))
                result__get_reverse_n_items = await i(db_request.get_reverse_n_items('key_3', 2))
                result__get_items_range = await i(db_request.get_items_range('key_1', 'key_3'))
                result__get_reverse_items_range = await i(db_request.get_reverse_items_range('key_3', 'key_1'))
                result__get_all_items = await i(db_request.get_all_items())
            
            self.assertEqual(result__get_first[1], 'value_1')
            self.assertEqual(result__get_last[1], 'value_3')
            self.assertEqual([item[0] for item in result__get_items.values()], ['value_1', 'value_2', 'value_3'], 'Result is wrong')
            self.assertEqual(result__get_n_items, [('key_1', 'value_1'), ('key_2', 'value_2')], 'Result is wrong')
            self.assertEqual(result__get_reverse_n_items, [('key_3', 'value_3'), ('key_2', 'value_2')], 'Result is wrong')
            self.assertEqual(result__get_items_range, [('key_1', 'value_1'), ('key_2', 'value_2'), ('key_3', 'value_3')], 'Result is wrong')
            self.assertEqual(result__get_reverse_items_range, [('key_3', 'value_3'), ('key_2', 'value_2'), ('key_1', 'value_1')], 'Result is wrong')
            self.assertEqual(result__get_all_items, [('key_1', 'value_1'), ('key_2', 'value_2'), ('key_3', 'value_3')], 'Result is wrong')

        with RT() as rt:
            run_in_loop(main, coro)
        
        print(f'Run time: {rt.rt}')

    # @unittest.skip('Skip')
    def test_07_delete_items(self):
        pifrl()
        async def coro(i: Interface, db_request: DbRequest) -> None:
            result__delete = None
            result__delete_kv = None
            result__delete_items = None
            result__delete_kv_items = None
            async with db(db_request, False) as db_service:
                await i(db_request.delete('key_1'))
                result__delete = await i(db_request.get_all_items())
                await i(db_request.delete_kv('key_2', 'value_2'))
                result__delete_kv = await i(db_request.get_all_items())
                await i(db_request.put_items({'key_4': 'value_4', 'key_5': 'value_5', 'key_6': 'value_6'}))
                result__put_items = await i(db_request.get_all_items())
                await i(db_request.delete_items({'key_3', 'key_4'}))
                result__delete_items = await i(db_request.get_all_items())
                await i(db_request.delete_kv_items({'key_5': 'value_5', 'key_6': 'value_6'}))
                result__delete_kv_items = await i(db_request.get_all_items())
            
            self.assertEqual(result__delete, [('key_2', 'value_2'), ('key_3', 'value_3')], 'Result is wrong')
            self.assertEqual(result__delete_kv, [('key_3', 'value_3')], 'Result is wrong')
            self.assertEqual(result__put_items, [('key_3', 'value_3'), ('key_4', 'value_4'), ('key_5', 'value_5'), ('key_6', 'value_6')], 'Result is wrong')
            self.assertEqual(result__delete_items, [('key_5', 'value_5'), ('key_6', 'value_6')], 'Result is wrong')
            self.assertEqual(result__delete_kv_items, list(), 'Result is wrong')

        with RT() as rt:
            run_in_loop(main, coro)
        
        print(f'Run time: {rt.rt}')

    # @unittest.skip('Skip')
    def test_08_cleanup(self):
        pifrl()
        async def coro(i: Interface, db_request: DbRequest) -> None:
            result = None
            async with db(db_request) as db_service:
                pass
            
            return result

        with RT() as rt:
            run_in_loop(main, coro)
        
        print(f'Run time: {rt.rt}')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import ctypes
import ctypes.wintypes as wintypes
from ctypes import windll
import shutil
from cengal.help_tools import current_line

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


_INSTALLING_BUFFER = {
    'copy': shutil.copy,
    'copy2': shutil.copy2,
}


LPDWORD = ctypes.POINTER(wintypes.DWORD)
LPOVERLAPPED = wintypes.LPVOID
LPSECURITY_ATTRIBUTES = wintypes.LPVOID

GENERIC_READ = 0x80000000
GENERIC_WRITE = 0x40000000
GENERIC_EXECUTE = 0x20000000
GENERIC_ALL = 0x10000000

FILE_SHARE_DELETE = 0x00000004
FILE_SHARE_READ = 0x00000001
FILE_SHARE_WRITE = 0x00000002

CREATE_NEW = 1
CREATE_ALWAYS = 2
OPEN_EXISTING = 3
OPEN_ALWAYS = 4
TRUNCATE_EXISTING = 5

FILE_ATTRIBUTE_NORMAL = 0x00000080

INVALID_HANDLE_VALUE = -1
INVALID_HANDLE_VALUE_CTYPES = wintypes.HANDLE(INVALID_HANDLE_VALUE).value

NULL = 0
FALSE = wintypes.BOOL(0)
TRUE = wintypes.BOOL(1)

WIN_UNICODE_PATH_PREFIX = '\\\\?\\'
WUPP = WIN_UNICODE_PATH_PREFIX


CreateFileW = None
if 'nt' == os.name:
    from cengal.ctypes_tools.libraries import kernel32
    from cengal.ctypes_tools.tools import cwfunc_def

    CreateFileW = cwfunc_def(kernel32, 'CreateFileW', (
            wintypes.LPWSTR,  # _In_ - LPCTSTR lpFileName
            wintypes.DWORD,  # _In_ - DWORD dwDesiredAccess
            wintypes.DWORD,  # _In_ - DWORD dwShareMode
            LPSECURITY_ATTRIBUTES,  # _In_opt_ - LPSECURITY_ATTRIBUTES lpSecurityAttributes
            wintypes.DWORD,  # _In_ - DWORD dwCreationDisposition
            wintypes.DWORD,  # _In_ - DWORD dwFlagsAndAttributes
            wintypes.HANDLE,  # _In_opt_ - HANDLE hTemplateFile
        ), wintypes.HANDLE)


def copy(src_name, dst_name):
    # ret = ctypes.windll.kernel32.CopyFileW(WUPP + src_name, WUPP + dst_name, False)
    ret = ctypes.windll.kernel32.CopyFileW(src_name, dst_name, False)
    if not ret:
        raise ctypes.WinError()

    # ret = ctypes.windll.kernel32.SetFileAttributesW(WUPP + dst_name, FILE_ATTRIBUTE_NORMAL)
    ret = ctypes.windll.kernel32.SetFileAttributesW(dst_name, FILE_ATTRIBUTE_NORMAL)
    if not ret:
        raise ctypes.WinError()

    # destination_file_handle = create_file(WUPP + dst_name, GENERIC_WRITE, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL)
    destination_file_handle = create_file(dst_name, GENERIC_WRITE, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL)
    if INVALID_HANDLE_VALUE_CTYPES == destination_file_handle.value:
        raise ctypes.WinError()

    try:
        creation_time = ctypes.wintypes.FILETIME()
        last_access_time = ctypes.wintypes.FILETIME()
        last_write_time = ctypes.wintypes.FILETIME()
        ret = ctypes.windll.kernel32.SetFileTime(destination_file_handle,
                                                 ctypes.byref(creation_time),
                                                 ctypes.byref(last_access_time),
                                                 ctypes.byref(last_write_time))
        if not ret:
            raise ctypes.WinError()
    except:
        raise
    finally:
        ret = ctypes.windll.kernel32.CloseHandle(destination_file_handle)
        if not ret:
            raise ctypes.WinError()


def copy2(src_name, dst_name):
    # ret = ctypes.windll.kernel32.CopyFileW(WUPP + src_name, WUPP + dst_name, False)
    ret = ctypes.windll.kernel32.CopyFileW(src_name, dst_name, False)
    if not ret:
        raise ctypes.WinError()


def create_file(filename, access, mode, security_attributes, creation, flags, template_file_handle):
    """See: CreateFile function
    http://msdn.microsoft.com/en-us/library/windows/desktop/aa363858(v=vs.85).aspx
    """
    return wintypes.HANDLE(CreateFileW(filename,
                                         access,
                                         mode,
                                         security_attributes,
                                         creation,
                                         flags,
                                         template_file_handle))


def install_module():
    shutil.copy = copy
    shutil.copy2 = copy2


def uninstall_module():
    shutil.copy = _INSTALLING_BUFFER['copy']
    shutil.copy2 = _INSTALLING_BUFFER['copy2']



__all__ = ['write_uint64', 'read_uint64', 'zero_memory']


from typing import Callable, Any


def write_uint64(base_address: int, offset: int, value: int) -> None: ...
def write_int64(base_address: int, offset: int, value: int) -> None: ...
def write_uint32(base_address: int, offset: int, value: int) -> None: ...
def write_int32(base_address: int, offset: int, value: int) -> None: ...
def write_uint16(base_address: int, offset: int, value: int) -> None: ...
def write_int16(base_address: int, offset: int, value: int) -> None: ...
def write_uint8(base_address: int, offset: int, value: int) -> None: ...
def write_int8(base_address: int, offset: int, value: int) -> None: ...
def write_float(base_address: int, offset: int, value: int) -> None: ...
def write_double(base_address: int, offset: int, value: int) -> None: ...


def read_uint64(base_address: int, offset: int) -> int: ...
def read_int64(base_address: int, offset: int) -> int: ...
def read_uint32(base_address: int, offset: int) -> int: ...
def read_int32(base_address: int, offset: int) -> int: ...
def read_uint16(base_address: int, offset: int) -> int: ...
def read_int16(base_address: int, offset: int) -> int: ...
def read_uint8(base_address: int, offset: int) -> int: ...
def read_int8(base_address: int, offset: int) -> int: ...
def read_float(base_address: int, offset: int) -> int: ...
def read_double(base_address: int, offset: int) -> int: ...


def zero_memory(base_address: int, offset: int, size: int) -> None: ...


def mask_least_significant_bits(hash_value: int, n: int) -> int: ...


def list__get_item(key: int, base_address: int, offset__pointer_to_internal_list: int, get_obj_callback: Callable) -> int: ...
def list__set_item(key: int, value: Any, base_address: int, offset__pointer_to_internal_list: int, put_obj_callback: Callable) -> int: ...



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['find_brackets', 'find_text_in_brackets', 'find_text_with_brackets', 'replace_text_in_brackets', 'replace_text_with_brackets']


from typing import Optional, Tuple, Callable, List
from cengal.text_processing.text_processing import Text, find_text, replace_slice, DEFAULT_ENCODING
from .brackets import *


def find_brackets(data: Text, brackets: BracketPair, start: int = 0, stop: Optional[int] = None, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Tuple[Optional[slice], Optional[slice]]:
    max_start = max(start, 0)
    max_start_slice = slice(max_start, max_start)
    end = len(data)
    stop = end if stop is None else stop
    min_end = min(end, stop)

    for l_bracket in brackets.left:
        if l_bracket.text():
            l_place = find_text(data, l_bracket.bracket_id, start, stop, encoding, normalizer)
        else:
            l_place = max_start_slice
        
        if l_place is not None:
            break
    
    for r_bracket in brackets.right:
        if r_bracket.text():
            r_place = find_text(data, r_bracket.bracket_id, (l_place or max_start_slice).stop, stop, encoding, normalizer)
        else:
            r_place = slice(min_end, min_end)
        
        if r_place is not None:
            break
    
    return l_place, r_place


def find_text_in_brackets(data: Text, brackets: BracketPair, start: int = 0, stop: Optional[int] = None, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Optional[slice]:
    l_place, r_place = find_brackets(data, brackets, start, stop, encoding, normalizer)
    if (l_place is None) or (r_place is None):
        return None
    
    return slice(l_place.stop, r_place.start)


def find_text_with_brackets(data: Text, brackets: BracketPair, start: int = 0, stop: Optional[int] = None, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Optional[slice]:
    l_place, r_place = find_brackets(data, brackets, start, stop, encoding, normalizer)
    if (l_place is None) or (r_place is None):
        return None
    
    return slice(l_place.start, r_place.stop)


def replace_text_in_brackets(data: Text, brackets: BracketPair, text: Text, count: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Tuple[Text, List[Tuple[slice, slice]]]:
    """_summary_

    Args:
        data (Text): _description_
        brackets (BracketPair): _description_
        text (Text): _description_
        count (int, optional): The same purpose and behaviour as in str().replace(), bytes().replace(), bytearray().replace(). Defaults to -1.
        encoding (Optional[str], optional): _description_. Defaults to DEFAULT_ENCODING.
        normalizer (Optional[Callable], optional): _description_. Defaults to None.

    Returns:
        Tuple[Text, List[Tuple[slice, slice]]]: Result text and a list with replacements logs (an old slice, a new slice)
    """
    result: List[slice] = list()
    parts: List[Text] = list()
    iter = None if count < 0 else count
    while (iter is None) or (iter > 0):
        old_text_slice = find_text_in_brackets(data, brackets, encoding=encoding, normalizer=normalizer)
        if old_text_slice is None:
            break

        data, new_text_slice = replace_slice(data, old_text_slice, text, encoding=encoding, normalizer=normalizer)
        result.append((old_text_slice, new_text_slice))
        parts.append(data[:new_text_slice.stop])
        data = data[new_text_slice.stop:]
        iter = None if count < 0 else iter - 1
    
    parts.append(data)
    data = type(data)().join(parts)
    return data, result


def replace_text_with_brackets(data: Text, brackets: BracketPair, text: Text, count: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Tuple[Text, List[Tuple[slice, slice]]]:
    """_summary_

    Args:
        data (Text): _description_
        brackets (BracketPair): _description_
        text (Text): _description_
        count (int, optional): The same purpose and behaviour as in str().replace(), bytes().replace(), bytearray().replace(). Defaults to -1.
        encoding (Optional[str], optional): _description_. Defaults to DEFAULT_ENCODING.
        normalizer (Optional[Callable], optional): _description_. Defaults to None.

    Returns:
        Tuple[Text, List[Tuple[slice, slice]]]: Result text and a list with replacements logs (an old slice, a new slice)
    """
    result: List[slice] = list()
    parts: List[Text] = list()
    iter = None if count < 0 else count
    while (iter is None) or (iter > 0):
        old_text_slice = find_text_with_brackets(data, brackets, encoding=encoding, normalizer=normalizer)
        if old_text_slice is None:
            break

        data, new_text_slice = replace_slice(data, old_text_slice, text, encoding=encoding, normalizer=normalizer)
        result.append((old_text_slice, new_text_slice))
        parts.append(data[:new_text_slice.stop])
        data = data[new_text_slice.stop:]
        iter = None if count < 0 else iter - 1

    parts.append(data)
    data = type(data)().join(parts)
    return data, result



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    'ServiceWithADirectRequestMixin', 
    'put_request_to_service',
    'try_put_request_to_service',
    'put_request_to_service_with_context', 
    'try_put_request_to_service_with_context', 
    'make_request_to_service_with_context', 
    'try_make_request_to_service_with_context', 
    'amake_request_to_service_with_context', 
    'atry_make_request_to_service_with_context', 
    'make_request_to_service', 
    'try_make_request_to_service', 
    'amake_request_to_service', 
    'atry_make_request_to_service'
]


from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.introspection.inspect import get_exception
from cengal.code_flow_control.smart_values import ValueExistence
from typing import Tuple, List, Optional, Any, Union, cast, Type

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class ServiceWithADirectRequestMixin:
    def _add_direct_request(self, *args, **kwargs) -> ValueExistence:
        raise NotImplementedError


def put_request_to_service_with_context(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], service_type: Type[ServiceWithADirectRequestMixin], *args, **kwargs) -> ValueExistence:
    loop, interface, coro_alive = context
    if loop is None:
        raise CoroSchedulerContextIsNotAvailable

    # Outside the loop or in the service
    service: ServiceWithADirectRequestMixin = loop.get_service_instance(service_type)
    return service._add_direct_request(*args, **kwargs)


def try_put_request_to_service_with_context(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], service_type: Type[ServiceWithADirectRequestMixin], *args, **kwargs) -> ValueExistence:
    loop, interface, coro_alive = context
    if loop is None:
        return None

    # Outside the loop or in the service
    service: ServiceWithADirectRequestMixin = loop.get_service_instance(service_type)
    return service._add_direct_request(*args, **kwargs)


def make_request_to_service_with_context(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], service_type: Type[ServiceWithADirectRequestMixin], *args, **kwargs) -> ValueExistence:
    loop, interface, coro_alive = context
    if loop is None:
        raise CoroSchedulerContextIsNotAvailable

    if coro_alive and isinstance(interface, InterfaceGreenlet):
        # In greenlet coroutine
        return True, interface(service_type, *args, **kwargs)
    else:
        # Outside the loop or in the service
        service: ServiceWithADirectRequestMixin = loop.get_service_instance(service_type)
        return service._add_direct_request(*args, **kwargs)


def try_make_request_to_service_with_context(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], service_type: Type[ServiceWithADirectRequestMixin], *args, **kwargs) -> ValueExistence:
    loop, interface, coro_alive = context
    if loop is None:
        return None

    if coro_alive and isinstance(interface, InterfaceGreenlet):
        # In greenlet coroutine
        return True, interface(service_type, *args, **kwargs)
    else:
        # Outside the loop or in the service
        service: ServiceWithADirectRequestMixin = loop.get_service_instance(service_type)
        return service._add_direct_request(*args, **kwargs)


async def amake_request_to_service_with_context(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], service_type: Type[ServiceWithADirectRequestMixin], *args, **kwargs) -> ValueExistence:
    loop, interface, coro_alive = context
    if loop is None:
        raise CoroSchedulerContextIsNotAvailable

    if coro_alive and isinstance(interface, InterfaceAsyncAwait):
        # In awaitable coroutine
        return True, await interface(service_type, *args, **kwargs)
    else:
        # Outside the loop or in the service
        service: ServiceWithADirectRequestMixin = loop.get_service_instance(service_type)
        return service._add_direct_request(*args, **kwargs)


async def atry_make_request_to_service_with_context(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], service_type: Type[ServiceWithADirectRequestMixin], *args, **kwargs) -> ValueExistence:
    loop, interface, coro_alive = context
    if loop is None:
        return None

    if coro_alive and isinstance(interface, InterfaceAsyncAwait):
        # In awaitable coroutine
        return True, await interface(service_type, *args, **kwargs)
    else:
        # Outside the loop or in the service
        service: ServiceWithADirectRequestMixin = loop.get_service_instance(service_type)
        return service._add_direct_request(*args, **kwargs)


def put_request_to_service(service_type: Type[ServiceWithADirectRequestMixin], *args, **kwargs) -> ValueExistence:
    return put_request_to_service_with_context(get_interface_and_loop_with_backup_loop(), service_type, *args, **kwargs)


def try_put_request_to_service(service_type: Type[ServiceWithADirectRequestMixin], *args, **kwargs) -> ValueExistence:
    return try_put_request_to_service_with_context(get_interface_and_loop_with_backup_loop(), service_type, *args, **kwargs)


def make_request_to_service(service_type: Type[ServiceWithADirectRequestMixin], *args, **kwargs) -> ValueExistence:
    return make_request_to_service_with_context(get_interface_and_loop_with_backup_loop(), service_type, *args, **kwargs)


def try_make_request_to_service(service_type: Type[ServiceWithADirectRequestMixin], *args, **kwargs) -> ValueExistence:
    return try_make_request_to_service_with_context(get_interface_and_loop_with_backup_loop(), service_type, *args, **kwargs)


async def amake_request_to_service(service_type: Type[ServiceWithADirectRequestMixin], *args, **kwargs) -> ValueExistence:
    return await amake_request_to_service_with_context(get_interface_and_loop_with_backup_loop(), service_type, *args, **kwargs)


async def atry_make_request_to_service(service_type: Type[ServiceWithADirectRequestMixin], *args, **kwargs) -> ValueExistence:
    return await atry_make_request_to_service_with_context(get_interface_and_loop_with_backup_loop(), service_type, *args, **kwargs)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['app_dir_path', 'app_dir_path_cached', 'AppDirPath', 'SHGetKnownFolderPathError']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


# More info:
# https://docs.microsoft.com/en-us/windows/win32/shell/knownfolderid
# https://docs.microsoft.com/en-us/windows/win32/api/shlobj_core/nf-shlobj_core-shgetknownfolderpath


import os
from ctypes import POINTER, byref, c_wchar_p
from cengal.ctypes_tools.types import GUID
from cengal.ctypes_tools.constants import KnownfolderidConstants
from cengal.ctypes_tools.result_api import SHGetKnownFolderPathError, SHGetKnownFolderPath
from cengal.file_system.directory_manager import ensure_dir as ensure_dir_exists
from .app_directory_types import AppDirectoryType
from .app_dir_path_base import AppDirPathBase, DirTypeMappingItem, DirTypeMappingItem, BaseDirID, DirNameOrPath, norm_dir_name_or_path
from typing import Optional, Tuple, Dict
from functools import lru_cache


dir_type_mapping: Dict[AppDirectoryType, Tuple[GUID, Optional[DirNameOrPath], Optional[DirNameOrPath]]] = {
    AppDirectoryType.local_data: (KnownfolderidConstants.FOLDERID_LocalAppData, None, 'data'),
    AppDirectoryType.local_cache: (KnownfolderidConstants.FOLDERID_LocalAppData, None, 'cache'),
    AppDirectoryType.local_config: (KnownfolderidConstants.FOLDERID_LocalAppData, None, 'config'),
    AppDirectoryType.local_log: (KnownfolderidConstants.FOLDERID_LocalAppData, None, 'log'),
    AppDirectoryType.local_temp: (KnownfolderidConstants.FOLDERID_LocalAppData, None, 'temp'),
    AppDirectoryType.local_runtime: (KnownfolderidConstants.FOLDERID_LocalAppData, None, 'runtime'),

    AppDirectoryType.local_low_data: (KnownfolderidConstants.FOLDERID_LocalAppDataLow, None, 'data'),
    AppDirectoryType.local_low_cache: (KnownfolderidConstants.FOLDERID_LocalAppDataLow, None, 'cache'),
    AppDirectoryType.local_low_config: (KnownfolderidConstants.FOLDERID_LocalAppDataLow, None, 'config'),
    AppDirectoryType.local_low_log: (KnownfolderidConstants.FOLDERID_LocalAppDataLow, None, 'log'),
    AppDirectoryType.local_low_temp: (KnownfolderidConstants.FOLDERID_LocalAppDataLow, None, 'temp'),
    AppDirectoryType.local_low_runtime: (KnownfolderidConstants.FOLDERID_LocalAppDataLow, None, 'runtime'),

    AppDirectoryType.roaming_data: (KnownfolderidConstants.FOLDERID_RoamingAppData, None, 'data'),
    AppDirectoryType.roaming_cache: (KnownfolderidConstants.FOLDERID_RoamingAppData, None, 'cache'),
    AppDirectoryType.roaming_config: (KnownfolderidConstants.FOLDERID_RoamingAppData, None, 'config'),
    AppDirectoryType.roaming_log: (KnownfolderidConstants.FOLDERID_RoamingAppData, None, 'log'),
    AppDirectoryType.roaming_temp: (KnownfolderidConstants.FOLDERID_RoamingAppData, None, 'temp'),
    AppDirectoryType.roaming_runtime: (KnownfolderidConstants.FOLDERID_RoamingAppData, None, 'runtime'),

    AppDirectoryType.program_data: (KnownfolderidConstants.FOLDERID_ProgramData, None, 'data'),
    AppDirectoryType.program_cache: (KnownfolderidConstants.FOLDERID_ProgramData, None, 'cache'),
    AppDirectoryType.program_config: (KnownfolderidConstants.FOLDERID_ProgramData, None, 'config'),
    AppDirectoryType.program_log: (KnownfolderidConstants.FOLDERID_ProgramData, None, 'log'),
    AppDirectoryType.program_temp: (KnownfolderidConstants.FOLDERID_ProgramData, None, 'temp'),
    AppDirectoryType.program_runtime: (KnownfolderidConstants.FOLDERID_ProgramData, None, 'runtime'),

    AppDirectoryType.program_files: (KnownfolderidConstants.FOLDERID_ProgramFiles, None, None),
    AppDirectoryType.program_files_common: (KnownfolderidConstants.FOLDERID_ProgramFilesCommon, None, None),

    AppDirectoryType.user_profile_data: (KnownfolderidConstants.FOLDERID_Profile, None, None),
    AppDirectoryType.user_profile_program_files: (KnownfolderidConstants.FOLDERID_UserProgramFiles, None, None),
    AppDirectoryType.user_profile_program_files_common: (KnownfolderidConstants.FOLDERID_UserProgramFilesCommon, None, None),

    AppDirectoryType.local_static_data: (KnownfolderidConstants.FOLDERID_LocalAppData, None, 'static_data'),
    AppDirectoryType.local_low_static_data: (KnownfolderidConstants.FOLDERID_LocalAppDataLow, None, 'static_data'),
    AppDirectoryType.roaming_static_data: (KnownfolderidConstants.FOLDERID_RoamingAppData, None, 'static_data'),
    AppDirectoryType.program_static_data: (KnownfolderidConstants.FOLDERID_ProgramData, None, 'static_data'),
}


def app_dir_path(dir_type: AppDirectoryType, app_name_or_path: Optional[DirNameOrPath] = None, with_structure: bool = True, ensure_dir: bool = True) -> str:
    if app_name_or_path is not None:
        app_name_or_path = str(app_name_or_path)
    
    if AppDirectoryType.user_profile_data == dir_type:
        if app_name_or_path is not None:
            app_name_or_path = '.' + norm_dir_name_or_path(app_name_or_path)

    mapping: Tuple[GUID, Optional[DirNameOrPath], Optional[DirNameOrPath]] = dir_type_mapping[dir_type]
    base_dir_path = SHGetKnownFolderPath(mapping[0])
    result_list = [base_dir_path]
    result_list.extend(norm_dir_name_or_path(mapping[1]))
    if app_name_or_path is not None:
        result_list.append(norm_dir_name_or_path(app_name_or_path))
        if with_structure:
            result_list.extend(norm_dir_name_or_path(mapping[2]))
    
    result_path = os.path.normpath(os.path.join(*result_list))
    if ensure_dir:
        ensure_dir_exists(result_path)
    
    return result_path


@lru_cache(maxsize=len(AppDirectoryType) * 2 * 10, typed=True)
def app_dir_path_cached(dir_type: AppDirectoryType, app_name_or_path: Optional[DirNameOrPath] = None, with_structure: bool = True, ensure_dir: bool = True):
    return app_dir_path(dir_type, app_name_or_path, with_structure, ensure_dir)


class AppDirPath(AppDirPathBase):
    def __init__(self, max_cache_size: Optional[int] = None) -> None:
        super().__init__(max_cache_size)
    
    def dir_type_mapping(self, dir_type: AppDirectoryType) -> DirTypeMappingItem:
        return dir_type_mapping[dir_type]
    
    def base_dir_id_to_path(self, base_dir_id: BaseDirID) -> str:
        return SHGetKnownFolderPath(base_dir_id)
